from typing import Dict, List, Union

from maestrowf.datastructures.core import ParameterGenerator


def get_custom_generator(env, **kwargs) -> ParameterGenerator:
    """
    Generates parameters to feed at the CL to hello_world.py.
    """
    p_gen: ParameterGenerator = ParameterGenerator()
    params: Dict[str, Union[List[float], str]] = {
        "X2": {"values": [1 / i for i in range(3, 6)], "label": "X2.%%"},
        "N_NEW": {"values": [2**i for i in range(1, 4)], "label": "N_NEW.%%"},
    }
    key: str
    value: Union[List[float], str]
    for key, value in params.items():
        p_gen.add_parameter(key, value["values"], value["label"])

    return p_gen



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module houses [`SampleIndex`][common.sample_index.SampleIndex] factory methods.
"""
from typing import List

from parse import parse

from merlin.common.sample_index import MAX_SAMPLE, SampleIndex
from merlin.utils import cd


# These pylint errors I've disabled are for "too many arguments"
# and "too many local variables". I think the functions are still clear
# pylint: disable=R0913,R0914


def create_hierarchy(
    num_samples: int,
    bundle_size: int,
    directory_sizes: List[int] = None,
    root: str = ".",
    start_sample_id: int = 0,
    start_bundle_id: int = 0,
    address: str = "",
    n_digits: int = 1,
) -> SampleIndex:
    """
    Factory method to create a [`SampleIndex`][common.sample_index.SampleIndex]
    hierarchy based on the number of samples.

    This method wraps the
    [`create_hierarchy_from_max_sample`][common.sample_index_factory.create_hierarchy_from_max_sample]
    function, which operates on a maximum sample basis rather than a total
    sample count.

    Args:
        num_samples (int): The total number of samples.
        bundle_size (int): The maximum number of samples a bundle file can handle.
        directory_sizes (List[int]): A list specifying the number of samples each directory
            is responsible for.
        root (str): The root path of the index.
        start_sample_id (int): The starting sample ID.
        start_bundle_id (int): The starting bundle ID.
        address (str): An optional address prefix for the hierarchy.
        n_digits (int): The number of digits to pad the directory names.

    Returns:
        (common.sample_index.SampleIndex): The root [`SampleIndex`][common.sample_index.SampleIndex]
            object representing the hierarchy.
    """
    if directory_sizes is None:
        directory_sizes = []
    return create_hierarchy_from_max_sample(
        num_samples + start_sample_id,
        bundle_size,
        directory_sizes=directory_sizes,
        root=root,
        start_bundle_id=start_bundle_id,
        min_sample=start_sample_id,
        address=address,
        n_digits=n_digits,
    )


def create_hierarchy_from_max_sample(
    max_sample: int,
    bundle_size: int,
    directory_sizes: List[int] = None,
    root: str = ".",
    start_bundle_id: int = 0,
    min_sample: int = 0,
    address: str = "",
    n_digits: int = 1,
) -> SampleIndex:
    """
    Constructs a [`SampleIndex`][common.sample_index.SampleIndex] hierarchy based on
    the maximum sample ID and chunking size at each depth.

    This method adds new [`SampleIndex`][common.sample_index.SampleIndex] objects as
    children if `directory_sizes` is provided.

    Args:
        max_sample: The maximum Sample ID this hierarchy is responsible for.
        bundle_size: The maximum number of samples a bundle file can handle.
        directory_sizes: A list specifying the number of samples each directory
            is responsible for.
        root: The root path of this index.
        start_bundle_id: The starting bundle ID.
        min_sample: The starting sample ID.
        address: An optional address prefix for the hierarchy.
        n_digits: The number of digits to pad the directory names.

    Returns:
        (common.sample_index.SampleIndex): The root
            [`SampleIndex`][common.sample_index.SampleIndex] object representing
            the constructed hierarchy.
    """
    if directory_sizes is None:
        directory_sizes = []

    # The dict of children nodes.
    children = {}
    # The child_id. Used for naming directory children.
    child_id = 0
    # The bundle_id - will increment as we add children.
    bundle_id = start_bundle_id
    # Number of samples the child node is responsible for.
    num_samples_per_child = directory_sizes[0] if directory_sizes else bundle_size

    if address:
        address_prefix = address + "."
    else:
        address_prefix = address

    for i in range(min_sample, max_sample, num_samples_per_child):
        child_min_sample_id = i
        child_max_sample_id = min(i + num_samples_per_child, max_sample)

        child_dir = f"{child_id}".zfill(n_digits)
        child_address = address_prefix + child_dir

        if directory_sizes:
            # Append an SampleIndex sub-hierarchy child.
            children[child_address] = create_hierarchy_from_max_sample(
                child_max_sample_id,
                bundle_size,
                directory_sizes[1:],
                root=child_dir,
                min_sample=child_min_sample_id,
                start_bundle_id=bundle_id,
                address=child_address,
                n_digits=n_digits,
            )

            bundle_id += children[child_address].num_bundles
        else:
            # Append a bundle file child.
            children[child_address] = SampleIndex(
                child_min_sample_id,
                child_max_sample_id,
                {},
                f"samples{child_min_sample_id}-{child_max_sample_id}.ext",
                leafid=bundle_id,
                address=child_address,
            )
            bundle_id += 1
        child_id += 1
    num_bundles = bundle_id - start_bundle_id
    return SampleIndex(min_sample, max_sample, children, root, num_bundles=num_bundles, address=address)


def read_hierarchy(path: str) -> SampleIndex:
    """
    Reads a hierarchy from a specified path and constructs a
    [`SampleIndex`][common.sample_index.SampleIndex].

    This function reads a file named "sample_index.txt" in the given path,
    parsing its contents to create a hierarchical structure of
    [`SampleIndex`][common.sample_index.SampleIndex] objects based on the
    information found in the file.

    Args:
        path: The directory path where the sample index file is located.

    Returns:
        (common.sample_index.SampleIndex): The root
            [`SampleIndex`][common.sample_index.SampleIndex] object representing
            the hierarchy read from the file.
    """
    children = {}
    min_sample = MAX_SAMPLE
    max_sample = -MAX_SAMPLE
    num_bundles = 0
    with cd(path):
        with open("sample_index.txt", "r") as _file:
            token = _file.readline()
            while token:
                parsed_token = parse("{type}:{ID}\tname:{name}\tsamples:[{min:d},{max:d})\n", token)
                if parsed_token["type"] == "DIR":
                    subhierarchy = read_hierarchy(parsed_token["name"])
                    subhierarchy.address = parsed_token["ID"]
                    num_bundles += subhierarchy.num_bundles
                    children[parsed_token["ID"]] = subhierarchy
                if parsed_token["type"] == "BUNDLE":
                    children[parsed_token["ID"]] = SampleIndex(
                        parsed_token["min"],
                        parsed_token["max"],
                        {},
                        parsed_token["name"],
                        address=parsed_token["ID"],
                    )
                    num_bundles += 1
                min_sample = min(min_sample, parsed_token["min"])
                max_sample = max(max_sample, parsed_token["max"])
                token = _file.readline()
    top_index = SampleIndex(min_sample, max_sample, children, path, leafid=-1, num_bundles=num_bundles)
    return top_index



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
SQLite-based backend infrastructure for the Merlin application.

This package provides all components necessary to persist and manage Merlin's core data
entities (such as studies, runs, and workers) using SQLite as the storage backend.
It includes a generic store base, specialized store classes for each entity type,
connection handling logic, and a complete backend implementation conforming to
Merlin's backend interface.

Modules:
    sqlite_backend: Implements the `ResultsBackend` interface using SQLite.
    sqlite_connection: Provides a context-managed SQLite connection with safe configuration.
    sqlite_store_base: Defines a generic base class for entity stores.
    sqlite_stores: Contains concrete SQLite store classes for Merlin models.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for restarting existing Merlin workflows.

This module defines the `RestartCommand` class, which provides functionality
to restart an existing Merlin workflow from a previously saved workspace.

The command verifies the provided workspace directory, locates the appropriate
provenance specification file (an expanded YAML spec), and re-initializes the
workflow study from that point. It supports running the restarted workflow
either locally or in a distributed mode.
"""

# pylint: disable=duplicate-code

import glob
import logging
import os
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from typing import List, Optional

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.config.configfile import initialize_config
from merlin.router import run_task_server
from merlin.study.study import MerlinStudy
from merlin.utils import verify_dirpath, verify_filepath


LOG = logging.getLogger("merlin")


class RestartCommand(CommandEntryPoint):
    """
    Handles `restart` CLI command for restarting existing workflows.

    Methods:
        add_parser: Adds the `restart` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `restart` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `restart` command parser will be added.
        """
        restart: ArgumentParser = subparsers.add_parser(
            "restart",
            help="Restart a workflow using an existing Merlin workspace.",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        restart.set_defaults(func=self.process_command)
        restart.add_argument("restart_dir", type=str, help="Path to an existing Merlin workspace directory")
        restart.add_argument(  # TODO should this just be boolean instead of store_const?
            "--local",
            action="store_const",
            dest="run_mode",
            const="local",
            default="distributed",
            help="Run locally instead of distributed",
        )

    def process_command(self, args: Namespace):
        """
        CLI command for restarting a study.

        This function handles the restart process by verifying the specified restart
        directory, locating a valid provenance specification file, and initiating
        the study from that point.

        Args:
            args: Parsed CLI arguments containing:\n
                - `restart_dir`: Path to the directory where the restart specifications are located.
                - `run_mode`: The mode for running the study (e.g., normal, dry-run).

        Raises:
            ValueError: If the `restart_dir` does not contain a valid provenance spec file or
                if multiple files match the specified pattern.
        """
        print(banner_small)
        restart_dir: str = verify_dirpath(args.restart_dir)
        filepath: str = os.path.join(args.restart_dir, "merlin_info", "*.expanded.yaml")
        possible_specs: Optional[List[str]] = glob.glob(filepath)
        if not possible_specs:  # len == 0
            raise ValueError(f"'{filepath}' does not match any provenance spec file to restart from.")
        if len(possible_specs) > 1:
            raise ValueError(f"'{filepath}' matches more than one provenance spec file to restart from.")
        filepath: str = verify_filepath(possible_specs[0])
        LOG.info(f"Restarting workflow at '{restart_dir}'")
        study: MerlinStudy = MerlinStudy(filepath, restart_dir=restart_dir)

        if args.run_mode == "local":
            initialize_config(local_mode=True)

        run_task_server(study, args.run_mode)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module defines the `TaskServerMonitor` abstract base class, which serves as a common interface
for monitoring task servers. Task servers are responsible for managing the execution of tasks and
workers in distributed systems, and this class provides an abstraction for monitoring their health
and progress.

The `TaskServerMonitor` class is intended to be subclassed for specific task server implementations
(e.g., Celery, TaskVine). Subclasses must implement all abstract methods to provide task server-specific
functionality, such as waiting for workers and checking task queues.
"""

from abc import ABC, abstractmethod
from typing import List

from merlin.db_scripts.entities.run_entity import RunEntity


class TaskServerMonitor(ABC):
    """
    Abstract base class for monitoring task servers. This class defines the interface
    for monitoring tasks and workers for a specific task server (e.g., Celery, TaskVine).

    Subclasses must implement all abstract methods to provide specific functionality
    for their respective task server.

    Methods:
        wait_for_workers: Wait for workers to start up.
        check_workers_processing: Check if any workers are still processing tasks.
        restart_worker: Restart a dead worker.
        run_worker_health_check: Check the health of workers and restart any that are dead.
        check_tasks: Abstract method to check the status of tasks in a workflow run.
            Must be implemented by subclasses.
    """

    @abstractmethod
    def wait_for_workers(self, workers: List[str], sleep: int):  # TODO should workers list be worker names or IDs?
        """
        Wait for workers to start up.

        Args:
            workers: A list of worker names or IDs to wait for.
            sleep: The interval (in seconds) between checks for worker availability.

        Raises:
            NoWorkersException: When workers don't start in (`self.sleep` * 10) seconds.
        """

    @abstractmethod
    def check_workers_processing(self, queues: List[str]) -> bool:
        """
        Check if any workers are still processing tasks.

        Args:
            queues: A list of queue names to check for active tasks.

        Returns:
            True if workers are processing tasks in the specified queues, False otherwise.
        """

    @abstractmethod
    def run_worker_health_check(self, workers: List[str]):
        """
        Checks the health of the workers provided and restarts any that are dead.

        Args:
            workers: A list of workers to check for worker health.
        """

    @abstractmethod
    def check_tasks(self, run: RunEntity) -> bool:
        """
        Check the status of tasks in the given workflow run.

        Args:
            run: A [`RunEntity`][db_scripts.entities.run_entity.RunEntity] instance representing
                the workflow run whose tasks are being monitored.

        Returns:
            True if tasks are active in the workflow, False otherwise.
        """



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing run entities within the Merlin database.

This module defines the `RunManager` class, which extends the generic
[`EntityManager`][db_scripts.entity_managers.entity_manager.EntityManager] base
class to provide CRUD operations for runs associated with studies, workspaces,
and queues. The manager coordinates with the study and logical worker managers
for consistent data handling during create and delete operations.
"""

from __future__ import annotations

import logging
from typing import Any, List

from merlin.db_scripts.data_models import RunModel
from merlin.db_scripts.entities.run_entity import RunEntity
from merlin.db_scripts.entity_managers.entity_manager import EntityManager
from merlin.exceptions import StudyNotFoundError, WorkerNotFoundError


LOG = logging.getLogger("merlin")

# Purposefully ignoring this pylint message as each entity will have different parameter requirements
# pylint: disable=arguments-differ,arguments-renamed


class RunManager(EntityManager[RunEntity, RunModel]):
    """
    Manager for run entities.

    This class handles creation, retrieval, updating, and deletion of runs in the database.
    It maintains consistency by coordinating with study and logical worker entities during
    operations such as run creation and deletion.

    Attributes:
        backend: The database backend used for storing run entities.
        db: Reference to the main Merlin database, allowing access to other entity managers
            such as studies and logical workers.

    Methods:
        create: Create a new run associated with a study and workspace.
        get: Retrieve a run by its ID or workspace identifier.
        get_all: Retrieve all runs from the database.
        delete: Delete a run and perform cleanup of related entities.
        delete_all: Delete all runs in the database.
        set_db_reference: Set the reference to the main Merlin database for cross-entity operations.
    """

    def create(self, study_name: str, workspace: str, queues: List[str], **kwargs: Any) -> RunEntity:
        """
        Create a new run associated with a study, workspace, and queues.

        This method ensures the study exists (creating it if necessary), then creates
        and saves a new run entity. Additional keyword arguments that correspond to valid
        [`RunModel`][db_scripts.data_models.RunModel] fields are included; other kwargs are
        stored as additional data.

        Args:
            study_name (str): Name of the study this run belongs to.
            workspace (str): Workspace identifier for the run.
            queues (List[str]): List of queues associated with the run.
            **kwargs (Any): Additional optional fields for the run entity.

        Returns:
            The created run entity.
        """
        # Create the study if it doesn't exist
        study_entity = self.db.studies.create(study_name)

        # Filter valid fields for the RunModel
        valid_fields = {f.name for f in RunModel.get_class_fields()}
        valid_kwargs = {key: val for key, val in kwargs.items() if key in valid_fields}
        additional_data = {key: val for key, val in kwargs.items() if key not in valid_fields}

        # Create the RunModel and save it
        new_run = RunModel(
            study_id=study_entity.get_id(),
            workspace=workspace,
            queues=queues,
            **valid_kwargs,
            additional_data=additional_data,
        )
        run_entity = RunEntity(new_run, self.backend)
        run_entity.save()

        # Add the run ID to the study
        study_entity.add_run(run_entity.get_id())

        return run_entity

    def get(self, run_id_or_workspace: str) -> RunEntity:
        """
        Retrieve a run entity by its unique ID or workspace identifier.

        Args:
            run_id_or_workspace (str): The unique identifier or workspace string of the run.

        Returns:
            The run entity corresponding to the given identifier.

        Raises:
            RunNotFoundError: If no run is found matching the identifier.
        """
        return self._get_entity(RunEntity, run_id_or_workspace)

    def get_all(self) -> List[RunEntity]:
        """
        Retrieve all run entities stored in the database.

        Returns:
            A list of all run entities.
        """
        return self._get_all_entities(RunEntity, "run")

    def delete(self, run_id_or_workspace: str):
        """
        Delete a run entity by its ID or workspace identifier.

        Performs cleanup operations to maintain consistency:\n
        - Removes the run from its associated study.
        - Removes the run from all logical workers referencing it.

        Args:
            run_id_or_workspace (str): The unique identifier or workspace string of the run to delete.

        Raises:
            RunNotFoundError: If no run is found matching the identifier.
        """

        def cleanup_run(run):
            # Remove from study
            try:
                study = self.db.studies.get(run.get_study_id())
                study.remove_run(run.get_id())
            except StudyNotFoundError:
                LOG.warning(f"Couldn't find study with id {run.get_study_id()}. Continuing with run delete.")

            # Remove from logical workers
            for worker_id in run.get_workers():
                try:
                    logical_worker = self.db.logical_workers.get(worker_id=worker_id)
                    logical_worker.remove_run(run.get_id())
                except WorkerNotFoundError:
                    LOG.warning(f"Couldn't find logical worker with id {worker_id}. Continuing with run delete.")

        self._delete_entity(RunEntity, run_id_or_workspace, cleanup_fn=cleanup_run)

    def delete_all(self):
        """
        Delete all run entities from the database.

        This method calls `delete` on each run entity to ensure proper cleanup.
        """
        self._delete_all_by_type(get_all_fn=self.get_all, delete_fn=self.delete, entity_name="runs")

    def set_db_reference(self, db: MerlinDatabase):  # noqa: F821  pylint: disable=undefined-variable
        """
        Set a reference to the main Merlin database object for cross-entity operations.

        This allows the manager to access other entity managers (e.g., logical workers) when
        performing operations like cleanup during deletions.

        Args:
            db (db_scripts.merlin_db.MerlinDatabase): The database object that provides
                access to related entity managers.
        """
        self.db = db



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module contains the logic for managing the sample hierarchy, including
the implementation of the [`SampleIndex`][common.sample_index.SampleIndex] class.
"""

import logging
import os
from contextlib import suppress
from typing import Callable, Dict, Generator, List, Tuple


LOG = logging.getLogger(__name__)

MAX_SAMPLE = 99999999999999999999999999999999999


def new_dir(path: str):
    """
    Create a new directory at the specified path if it does not already exist.

    This function attempts to create the directory and suppresses any
    OSError that may occur if the directory already exists.

    Args:
        path: The path where the new directory should be created.
    """
    with suppress(OSError):
        os.makedirs(path)


def uniform_directories(num_samples: int = MAX_SAMPLE, bundle_size: int = 1, level_max_dirs: int = 100) -> List[int]:
    """
    Create a directory hierarchy with uniformly increasing directory sizes.

    This function generates a list of directory sizes, starting from
    the specified `bundle_size` and increasing by a factor of
    `level_max_dirs` until the total number of samples is reached.

    Args:
        num_samples: The total number of samples to consider.
        bundle_size: The initial size of each bundle.
        level_max_dirs: The factor by which to increase directory sizes at each level.

    Returns:
        A list of integers representing the sizes of directories in the hierarchy.
    """
    directory_sizes = [bundle_size]
    while directory_sizes[0] < num_samples:
        directory_sizes.insert(0, directory_sizes[0] * level_max_dirs)
    # We've gone over the total number of samples, remove the first entry
    del directory_sizes[0]
    return directory_sizes


class SampleIndex:
    """
    Represents a hierarchical structure for managing bundles of sample files.

    A [`SampleIndex`][common.sample_index.SampleIndex] serves as an in-situ
    representation of a directory hierarchy where bundles of sample result files
    are stored. This class provides factory methods to create complete hierarchies
    and to read an index from a previously stored file on disk.

    Attributes:
        address (str): The full address of this node.
        children (Dict[str, SampleIndex]): A dictionary containing the direct children of this node, which are also
            of type SampleIndex.
        depth (int): (class attribute) A class variable indicating the current depth in the
            hierarchy, primarily used for pretty printing.
        is_leaf (bool): Returns whether this node is a leaf in the hierarchy.
        is_directory (bool): Returns whether this node is a directory (not a leaf).
        is_parent_of_leaf (bool): Returns whether this node is the direct parent of a leaf.
        is_grandparent_of_leaf (bool): Returns whether this node is the parent of a parent of a leaf.
        is_great_grandparent_of_leaf (bool): Returns whether this node is the parent of a grandparent of a leaf.
        leafid (int): The unique leaf ID of this node.
        max (int): The maximum global sample ID of any child node.
        min (int): The minimum global sample ID of any child node.
        name (str): The name of this node in the hierarchy.
        num_bundles (int): The total number of bundles in this index.

    Methods:
        traverse: Yields the full path and associated node for each node meeting a specified condition.
        traverse_all: Returns a generator that traverses all nodes in the
            [`SampleIndex`][common.sample_index.SampleIndex].
        traverse_bundles: Returns a generator that traverses all bundles (leaves) in the
            [`SampleIndex`][common.sample_index.SampleIndex].
        traverse_directories: Returns a generator that traverses all directories in the
            [`SampleIndex`][common.sample_index.SampleIndex].
        check_valid_addresses_for_insertion: Validates addresses for insertion into the hierarchy.
        __getitem__: Retrieves a child node by its full address.
        __setitem__: Sets a child node at a specified full address.
        write_directory: Creates a new directory associated with this node.
        write_directories: Recursively writes directories for this node and its children.
        get_path_to_sample: Retrieves the file path to the bundle file with a specified sample ID.
        write_single_sample_index_file: Writes a single sample index file for this node.
        write_multiple_sample_index_files: Writes multiple sample index files for this node and its children.
        make_directory_string: Creates a delimited string representation of the directories in the index.
        __str__: Returns a string representation of the [`SampleIndex`][common.sample_index.SampleIndex],
            including its children.
    """

    # Class variable to indicate depth (mostly used for pretty printing).
    depth: int = -1

    def __init__(
        self,
        minid: int,
        maxid: int,
        children: Dict[str, "SampleIndex"],
        name: str,
        leafid: int = -1,
        num_bundles: int = 0,
        address: str = "",
    ):  # pylint: disable=R0913
        """
        Initializes a new instance of the `SampleIndex` class.

        Args:
            minid: The minimum global sample ID of any child node.
            maxid: The maximum global sample ID of any child node.
            children: A dictionary containing the direct children of this node,
                where the keys are the children's full addresses and the values are `SampleIndex` instances.
            name: The name of this node in the hierarchy.
            leafid: The unique leaf ID of this node.
            num_bundles: The total number of bundles in this index.
            address: The full address of this node.

        Raises:
            TypeError: If `children` is not of type `dict`.
        """

        # The direct children of this node, generally also of type SampleIndex.
        # A dictionary keyed by the childrens' full addresses.
        if not isinstance(children, dict):
            LOG.error("SampleIndex children must be of type dict")
            raise TypeError

        self.children: Dict[str, "SampleIndex"] = children
        self.name: str = name  # The name of this node.
        self.address: str = str(address)  # The address of this node

        # @NOTE: The following are only valid if no insertions,splits, or
        # joins have taken place since the last renumber() call, or this index
        # was constructed with apriori global knowledge of its contents.

        # The minimum global sample ID of any of the children of this node.
        self.min: int = minid

        # The maximum global sample ID of any of the children of this node.
        self.max: int = maxid

        # The total number of bundles in this index.
        self.num_bundles: int = num_bundles

        # The unique leaf ID of this node
        self.leafid: int = leafid

    @property
    def is_leaf(self) -> bool:
        """
        Indicates whether this node is a leaf in the hierarchy.

        A leaf node is defined as a node that has no children. This property
        returns True if the node has no children, and False otherwise.
        """
        return len(self.children) == 0

    @property
    def is_directory(self) -> bool:
        """
        Indicates whether this node is a directory (not a leaf).

        A directory node is defined as a node that has one or more children.
        This property returns True if the node has children, and False if it
        is a leaf node.
        """
        return len(self.children) > 0

    @property
    def is_parent_of_leaf(self) -> bool:
        """
        Indicates whether this node is the direct parent of a leaf.

        This property checks if the current node is a directory and if any of
        its children are leaf nodes. It returns True if at least one child is
        a leaf, and False otherwise.
        """
        if not self.is_directory:
            return False
        for child_val in self.children.values():
            if child_val.is_leaf:
                return True
        return False

    @property
    def is_grandparent_of_leaf(self) -> bool:
        """
        Indicates whether this node is the parent of a parent of a leaf.

        This property checks if the current node is a directory and if any of
        its children are parents of leaf nodes. It returns True if at least
        one child is a parent of a leaf, and False otherwise.
        """
        if not self.is_directory:
            return False
        for child_val in self.children.values():
            if child_val.is_parent_of_leaf:
                return True
        return False

    @property
    def is_great_grandparent_of_leaf(self) -> bool:
        """
        Indicates whether this node is the parent of a grandparent of a leaf.

        This property checks if the current node is a directory and if any of
        its children are grandparents of leaf nodes. It returns True if at
        least one child is a grandparent of a leaf, and False otherwise.
        """
        if not self.is_directory:
            return False
        for child_val in self.children.values():
            if child_val.is_grandparent_of_leaf:
                return True
        return False

    def traverse(
        self,
        path: str = None,
        conditional: Callable = lambda c: True,
        bottom_up: bool = True,
        top_level: bool = True,
    ) -> Generator[Tuple[str, "SampleIndex"], None, None]:
        """
        Traverse the tree structure and yield the full path and associated node
        for each node that meets the specified conditional criteria.

        This method allows for flexible traversal of the tree, either from
        the top down or bottom up, depending on the `bottom_up` parameter.
        Nodes are yielded based on whether they satisfy the provided
        conditional function.

        Notes:
            - The method uses a "SKIP ME" placeholder to manage the
            recursion flow, ensuring that the traversal can skip
            non-qualifying nodes without breaking the iteration.

        Args:
            path: The current path to this node. If None, it defaults to
                the name of the node.
            conditional: A function that takes a
                [`SampleIndex`][common.sample_index.SampleIndex] instance as its
                only argument and returns a boolean. It determines whether a
                node should be yielded.
            bottom_up: If True, yields leaf nodes first, moving upwards through
                the tree. If False, yields top-level nodes first.
            top_level: A flag used internally to control filtering of yielded
                values based on the conditional. Should only be set to False
                for recursive calls.

        Yields:
            A tuple containing the full path to the node and the
                associated [`SampleIndex`][common.sample_index.SampleIndex] node
                that meets the conditional criteria.
        """
        if path is None:
            path = self.name

        if not bottom_up and conditional(self):
            yield path, self

        for child_val in self.children.values():
            child_path = os.path.join(path, child_val.name)
            for node in child_val.traverse(child_path, conditional, bottom_up=bottom_up, top_level=False):
                if node != "SKIP ME":
                    yield node

        if bottom_up and conditional(self):
            yield path, self

        # Always yield something, unless this is the end of the recursion.
        if not top_level:
            yield "SKIP ME"

    def traverse_all(self, bottom_up: bool = True) -> Generator[Tuple[str, "SampleIndex"], None, None]:
        """
        Traverse all nodes in the [`SampleIndex`][common.sample_index.SampleIndex].

        This method returns a generator that yields all nodes in the
        [`SampleIndex`][common.sample_index.SampleIndex], regardless of their type
        (leaf or directory). The traversal order can be controlled by the `bottom_up`
        parameter.

        Notes:
            This method calls the [`traverse`][common.sample_index.SampleIndex.traverse]
            method with a conditional that always returns True, ensuring all nodes are
            included.

        Args:
            bottom_up: If True, yields leaf nodes first, moving upwards
                through the tree. If False, yields top-level nodes first.

        Returns:
            A tuple containing the full path to each node and the associated
                [`SampleIndex`][common.sample_index.SampleIndex] node.
        """
        return self.traverse(path=self.name, conditional=lambda c: True, bottom_up=bottom_up)

    def traverse_bundles(self) -> Generator[Tuple[str, "SampleIndex"], None, None]:
        """
        Traverse all Bundles (leaf nodes) in the [`SampleIndex`][common.sample_index.SampleIndex].

        This method returns a generator that yields only the leaf nodes
        (Bundles) in the [`SampleIndex`][common.sample_index.SampleIndex].
        It filters the nodes based on their type, ensuring that only leaves are returned.

        Notes:
            This method calls the [`traverse`][common.sample_index.SampleIndex.traverse]
            method with a conditional that checks if a node is a leaf, ensuring only Bundles
            are yielded.

        Returns:
            A tuple containing the full path to each Bundle and the associated
                [`SampleIndex`][common.sample_index.SampleIndex] node.
        """
        return self.traverse(path=self.name, conditional=lambda c: c.is_leaf)

    def traverse_directories(self, bottom_up: bool = False) -> Generator[Tuple[str, "SampleIndex"], None, None]:
        """
        Traverse all Directories in the [`SampleIndex`][common.sample_index.SampleIndex].

        This method returns a generator that yields all directory nodes
        in the [`SampleIndex`][common.sample_index.SampleIndex]. The
        traversal order can be controlled by the `bottom_up` parameter.

        Notes:
            This method calls the [`traverse`][common.sample_index.SampleIndex.traverse]
            method with a conditional that checks if a node is a directory, ensuring only
            directories are yielded.

        Args:
            bottom_up: If True, yields leaf directories first, moving
                upwards through the tree. If False, yields top-level
                directories first.

        Yields:
            A tuple containing the full path to each Directory and the
                associated [`SampleIndex`][common.sample_index.SampleIndex]
                node.
        """
        return self.traverse(path=self.name, conditional=lambda c: c.is_directory, bottom_up=bottom_up)

    @staticmethod
    def check_valid_addresses_for_insertion(full_address: str, sub_tree: "SampleIndex"):
        """
        Check if the provided address is valid for insertion into the subtree.

        This method traverses all nodes in the given subtree and verifies
        that no existing node's address conflicts with the specified
        `full_address`. If any node's address starts with the `full_address`,
        a TypeError is raised, indicating that the insertion would create
        an invalid state.

        Args:
            full_address: The full address to be checked for validity before
                insertion.
            sub_tree: The subtree in which the address will be checked.

        Raises:
            TypeError: If any node in the subtree has an address that
                conflicts with the `full_address`.
        """
        for _, node in sub_tree.traverse_all():
            if node.address[0 : len(full_address)] != full_address:
                raise TypeError

    def __getitem__(self, full_address: str) -> "SampleIndex":
        """
        Retrieve the subtree associated with the given full address.

        This method allows for accessing nodes in the
        [`SampleIndex`][common.sample_index.SampleIndex] using
        the full address. If the full address matches the current node's
        address, the node itself is returned. Otherwise, it recursively
        searches through the children to find the corresponding node.

        Args:
            full_address: The full address of the node to retrieve.

        Returns:
            The node associated with the specified full address.

        Raises:
            KeyError: If no node with the specified full address exists
                in the [`SampleIndex`][common.sample_index.SampleIndex].
        """
        if full_address == self.address:
            return self
        for child_val in list(self.children.values()):
            if full_address[0 : len(child_val.address)] == child_val.address:
                return child_val[full_address]
        raise KeyError

    def __setitem__(self, full_address: str, sub_tree: "SampleIndex"):
        """
        Set or replace the subtree associated with the given full address.

        This method allows for inserting or updating a node in the
        [`SampleIndex`][common.sample_index.SampleIndex]. If the full
        address matches the current node's address, a KeyError is raised
        to prevent self-assignment. The method searches through the children
        to find the appropriate location for insertion. If a node already
        exists at the specified address, it will be replaced after validating
        the insertion.

        Args:
            full_address: The full address of the node to set or replace.
            sub_tree: The subtree to insert or replace at the specified
                address.

        Raises:
            KeyError: If the full address matches the current node's
                address or if no matching child node is found for
                insertion.
        """
        if full_address == self.address:
            # This should never happen.
            raise KeyError

        delete_me = None
        for child_val in list(self.children.values()):
            if full_address[0 : len(child_val.address)] == child_val.address:
                if child_val.address == full_address:
                    delete_me = full_address
                    break

                child_val[full_address] = sub_tree
                return

        # Replace if we already have something at this address.
        if delete_me is not None:
            SampleIndex.check_valid_addresses_for_insertion(full_address, sub_tree)
            self.children.__delitem__(full_address)
            self.children[full_address] = sub_tree
            return
        raise KeyError

    def write_directory(self, path: str):
        """
        Create a new directory associated with this node in the graph.

        This method checks if the current node is a directory and, if so,
        creates a new directory at the specified path using the node's
        name. The directory is created using the
        [`new_dir`][common.sample_index.new_dir] function.

        Args:
            path: The base path where the new directory will be created.
        """
        if self.is_directory:
            new_dir(os.path.join(path, self.name))

    def write_directories(self, path: str = "."):
        """
        Create the directory tree associated with this node and its children.

        This method initiates the creation of the directory structure
        starting from the current node. It first creates the directory for
        the current node and then recursively creates directories for all
        child nodes. The base path can be specified, and the directories
        will be created relative to this path.

        Args:
            path: The base path where the directory tree will be created.
                Defaults to the current directory ("."), meaning the
                directories will be created in the current working
                directory.
        """
        self.write_directory(path)
        for child_val in list(self.children.values()):
            child_val.write_directories(os.path.join(path, self.name))

    def get_path_to_sample(self, sample_id: int) -> str:
        """
        Retrieve the file path to the bundle file associated with the specified
        sample ID.

        This method constructs the path to the bundle file by traversing the
        directory structure represented by the current node and its children.
        It checks each child node to determine if the provided `sample_id`
        falls within the range defined by the child's `min` and `max` attributes.
        If a matching child is found, the method recursively calls itself to
        build the complete path.

        Notes:
            - This method assumes that the global numbering system is known
            and that the `min` and `max` attributes of child nodes are
            correctly defined.
            - If no matching child is found, the method will return the
            current node's name as the base path.

        Args:
            sample_id: The identifier of the sample for which the file path is
                to be retrieved. This ID must correspond to a sample within the
                known global numbering system.

        Returns:
            The constructed file path to the bundle file associated
                with the specified `sample_id`.
        """
        path = self.name
        for child_val in self.children.values():
            if child_val.min <= sample_id < child_val.max:
                path = os.path.join(path, child_val.get_path_to_sample(sample_id))
        return path

    def write_single_sample_index_file(self, path: str) -> str:
        """
        Write the index file associated with this node.

        This method creates an index file named "sample_index.txt" in the
        specified directory path. The index file contains information about
        the child nodes of the current node. For each child, it records
        whether the child is a leaf or a directory, along with its address,
        name, and the range of sample IDs it covers.

        Notes:
            - This method will only execute if the current node is identified
            as a directory
            ([`self.is_directory`][common.sample_index.SampleIndex.is_directory]).
            - The index file format includes lines for each child in the
            following format:
                - For leaf nodes: `BUNDLE:<address>\tname:<name>\tSAMPLES:[<min>, <max>)`
                - For directory nodes: `DIR:<address>\tname:<name>\tSAMPLES:[<min>, <max>)`

        Args:
            path: The base path where the index file will be created.

        Returns:
            The full path to the created index file if the current node is
                a directory; otherwise, returns None.
        """
        if not self.is_directory:
            return None

        fname = os.path.join(path, self.name, "sample_index.txt")
        with open(fname, "w") as _file:
            for child_val in self.children.values():
                if child_val.is_leaf:
                    _file.write(
                        f"BUNDLE:{child_val.address}\tname:{child_val.name}\tSAMPLES:[{child_val.min}, {child_val.max})\n"
                    )
                else:
                    _file.write(
                        f"DIR:{child_val.address}\tname:{child_val.name}\tSAMPLES:[{child_val.min}, {child_val.max})\n"
                    )
        return fname

    def write_multiple_sample_index_files(self, path: str = ".") -> List[str]:
        """
        Write index files that correspond to the location in the directory hierarchy.

        This method generates index files for the current node and all its
        children, allowing for a structured representation of the sample
        indices in the directory hierarchy. It first writes a single index
        file for the current node and then recursively writes index files
        for each child node.

        Args:
            path: The base path where the index files will be created.
                Defaults to the current directory ("."), meaning the index
                files will be created in the current working directory.

        Returns:
            A list of file paths to the created index files.
        """
        filepath = self.write_single_sample_index_file(path)
        filepaths = []
        if filepath is not None:
            filepaths.append(filepath)
        for child_val in self.children.values():
            filepaths += child_val.write_multiple_sample_index_files(os.path.join(path, self.name))
        return filepaths

    def make_directory_string(self, delimiter: str = " ", just_leaf_directories: bool = True) -> str:
        """
        Create a delimited string representation of the directories in the index.

        This method generates a string that lists the directories in the
        current index, separated by the specified delimiter. The user can
        choose to include only the leaf directories or all directories.

        Notes:
            - The method utilizes the
            [`traverse_directories`][common.sample_index.SampleIndex.traverse_directories]
            function to retrieve the directory paths and their corresponding nodes.

        Args:
            delimiter: The characters used to separate the directories in the
                resulting string.
            just_leaf_directories: If True, only leaf directories (the bottom-level
                directories) will be included in the output. If False, all directories
                will be included.

        Returns:
            A string representation of the directories, formatted as a delimited
                list. For example: "0/0 0/1 0/2 1/0 1/1 1/2".
        """
        # fmt: off
        if just_leaf_directories:
            return delimiter.join(
                [
                    path for path, node in self.traverse_directories() if node.is_parent_of_leaf
                ]
            )
        # fmt: on
        return delimiter.join([path for path, _ in self.traverse_directories()])

    def __str__(self) -> str:
        """
        Return a string representation of the
        [`SampleIndex`][common.sample_index.SampleIndex] object.

        This method provides a formatted string that represents the current
        node in the sample index, including its address, type (BUNDLE or
        DIRECTORY), and relevant attributes such as minimum and maximum
        sample IDs. If the node is a directory, it recursively includes
        the string representations of its child nodes.

        Notes:
            - The method uses a class variable `depth` to manage indentation
            levels for nested directories, enhancing readability.
            - The output format varies depending on whether the node is a
            leaf or a directory.

        Returns:
            A formatted string representation of the
                [`SampleIndex`][common.sample_index.SampleIndex] object,
                including its children if applicable.
        """
        SampleIndex.depth = SampleIndex.depth + 1
        if self.is_leaf:
            result = ("   " * SampleIndex.depth) + f"{self.address}: BUNDLE {self.leafid} MIN {self.min} MAX {self.max}\n"
        else:
            result = (
                "   " * SampleIndex.depth
            ) + f"{self.address}: DIRECTORY MIN {self.min} MAX {self.max} NUM_BUNDLES {self.num_bundles}\n"
            for child_val in self.children.values():
                result += str(child_val)
        SampleIndex.depth = SampleIndex.depth - 1
        return result



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""Main functions for instantiating and running Merlin server containers."""

import logging
import os
import socket
import subprocess
import time
from argparse import Namespace

from merlin.server.server_config import (
    ServerStatus,
    config_merlin_server,
    create_server_config,
    dump_process_file,
    get_server_status,
    parse_redis_output,
    pull_process_file,
    pull_server_config,
    pull_server_image,
)
from merlin.server.server_util import AppYaml, RedisConfig, RedisUsers, ServerConfig


LOG = logging.getLogger("merlin")


def init_server():
    """
    Initializes the Merlin server by setting up the main configuration directory
    and local server configuration.
    """

    if not create_server_config():
        LOG.info("Merlin server initialization failed.")
        return
    pull_server_image()

    config_merlin_server()

    LOG.info("Merlin server initialization successful.")


def apply_config_changes(server_config: ServerConfig, args: Namespace):
    """
    Applies configuration changes to the Merlin server based on user-provided arguments.

    This function modifies the Redis configuration and user settings as specified by
    the user through the provided arguments. It updates various Redis settings, such as
    IP address, port, password, directory, snapshot settings, append mode, and file paths.
    If any changes are made, the updated configuration is written to the appropriate files.

    Args:
        server_config (server.server_util.ServerConfig): An instance of `ServerConfig` containing
            all the necessary configuration values for the server.
        args (Namespace): An argparse `Namespace` object containing user-provided arguments
            from the argument parser.
    """
    redis_config = RedisConfig(server_config.container.get_config_path())

    redis_config.set_ip_address(args.ipaddress)

    redis_config.set_port(args.port)

    redis_config.set_password(args.password)
    if args.password is not None:
        redis_users = RedisUsers(server_config.container.get_user_file_path())
        redis_users.set_password("default", args.password)
        redis_users.write()
        pass_file = server_config.container.get_pass_file_path()
        with open(pass_file, "w") as pfile:
            pfile.write(args.password)

    redis_config.set_directory(args.directory)

    redis_config.set_snapshot(seconds=args.snapshot_seconds, changes=args.snapshot_changes)

    redis_config.set_snapshot_file(args.snapshot_file)

    redis_config.set_append_mode(args.append_mode)

    redis_config.set_append_file(args.append_file)

    if redis_config.changes_made():
        redis_config.write()
        LOG.info("Merlin server config has changed. Restart merlin server to apply new configuration.")
        LOG.info("Run 'merlin server restart' to restart running merlin server")
        LOG.info("Run 'merlin server start' to start merlin server instance.")
    else:
        LOG.info("Add changes to config file and exisiting containers.")


# Pylint complains that there's too many branches in this function but
# it looks clean to me so we'll ignore it
def config_server(args: Namespace):  # pylint: disable=R0912
    """
    Processes the Merlin server configuration flags to make changes and edits
    to appropriate configurations based on user input.

    Args:
        args (Namespace): An argparse `Namespace` object containing user-provided arguments
            from the argument parser.
    """
    server_config_before_changes = pull_server_config()
    if not server_config_before_changes:
        LOG.error('Try to run "merlin server init" again to reinitialize values.')
        return False

    apply_config_changes(server_config_before_changes, args)

    server_config_after_changes = pull_server_config()
    if not server_config_after_changes:
        LOG.error('Try to run "merlin server init" again to reinitialize values.')
        return False

    # Read the user from the list of avaliable users
    redis_users = RedisUsers(server_config_after_changes.container.get_user_file_path())
    redis_config = RedisConfig(server_config_after_changes.container.get_config_path())

    if args.add_user is not None:
        # Log the user in a file
        if redis_users.add_user(user=args.add_user[0], password=args.add_user[1]):
            redis_users.write()
            LOG.info(f"Added user {args.add_user[0]} to merlin server")
            # Create a new user in container
            if get_server_status() == ServerStatus.RUNNING:
                LOG.info("Adding user to current merlin server instance")
                redis_users.apply_to_redis(redis_config.get_ip_address(), redis_config.get_port(), redis_config.get_password())
        else:
            LOG.error(f"User '{args.add_user[0]}' already exisits within current users")

    if args.remove_user is not None:
        # Remove user from file
        if redis_users.remove_user(args.remove_user):
            redis_users.write()
            LOG.info(f"Removed user {args.remove_user} to merlin server")
            # Remove user from container
            if get_server_status() == ServerStatus.RUNNING:
                LOG.info("Removing user to current merlin server instance")
                redis_users.apply_to_redis(redis_config.get_ip_address(), redis_config.get_port(), redis_config.get_password())
        else:
            LOG.error(f"User '{args.remove_user}' doesn't exist within current users.")

    return None


def status_server():
    """
    Retrieves and displays the current status of the Merlin server.

    This function checks the status of any running containers for the Merlin server
    and logs appropriate messages based on the server's state.
    """
    current_status = get_server_status()
    if current_status == ServerStatus.NOT_INITIALIZED:
        LOG.info("Merlin server has not been initialized.")
        LOG.info("Please initalize server by running 'merlin server init'")
    elif current_status == ServerStatus.MISSING_CONTAINER:
        LOG.info("Unable to find server image.")
        LOG.info("Ensure there is a .sif file in merlin server directory.")
    elif current_status == ServerStatus.NOT_RUNNING:
        LOG.info("Merlin server is not running.")
    elif current_status == ServerStatus.RUNNING:
        LOG.info("Merlin server is running.")


def check_for_not_running_server() -> bool:
    """
    Checks if the Merlin server status is `NOT_RUNNING` before starting a new server.

    If the server status is anything other than `NOT_RUNNING`, logs an appropriate
    error message to inform the user.

    Returns:
        True if the server status is `NOT_RUNNING`, False otherwise.
    """
    current_status = get_server_status()
    uninitialized_err = "Merlin server has not been intitialized. Please run 'merlin server init' first."
    status_errors = {
        ServerStatus.NOT_INITIALIZED: uninitialized_err,
        ServerStatus.MISSING_CONTAINER: uninitialized_err,
        ServerStatus.RUNNING: """Merlin server already running.
                              Stop current server with 'merlin server stop' before attempting to start a new server.""",
    }

    if current_status in status_errors:
        LOG.info(status_errors[current_status])
        return False

    return True


def start_container(server_config: ServerConfig) -> subprocess.Popen:
    """
    Starts a container based on the provided server configuration.

    This function uses the server configuration to locate the necessary image and
    configuration files, validates their existence, and starts the container using
    a subprocess.

    Args:
        server_config (server.server_util.ServerConfig): An instance of `ServerConfig`
            containing information about the server to start, including paths to the image
            and configuration files.

    Returns:
        A subprocess object representing the running container process, or `None` if
            required files are missing.
    """
    image_path = server_config.container.get_image_path()
    config_path = server_config.container.get_config_path()
    path_errors = {
        image_path: "image",
        config_path: "config file",
    }

    for path in (image_path, config_path):
        if not os.path.exists(path):
            LOG.error(f"Unable to find {path_errors[path]} at {path}")
            return None

    # Pylint wants us to use with here but we don't need that
    process = subprocess.Popen(  # pylint: disable=R1732
        server_config.container_format.get_run_command()
        .strip("\\")
        .format(
            command=server_config.container_format.get_command(),
            home_dir=server_config.container.get_config_dir(),
            image=image_path,
            config=config_path,
        )
        .split(),
        start_new_session=True,
        close_fds=True,
        stdout=subprocess.PIPE,
    )

    time.sleep(1)

    return process


def server_started(process: subprocess.Popen, server_config: ServerConfig) -> bool:
    """
    Verifies that the server started by [`start_container`][server.server_commands.start_container]
    is running properly.

    This function checks the Redis output to ensure the server started successfully,
    creates a process file for the container, and validates that the server status
    is `RUNNING`.

    Args:
        process (subprocess.Popen): The subprocess object representing the container
            process started by `start_container`.
        server_config (server.server_util.ServerConfig): An instance of `ServerConfig`
            containing information about the Redis server configuration.

    Returns:
        True if the server started successfully, False otherwise.
    """
    redis_start, redis_out = parse_redis_output(process.stdout)

    if not redis_start:
        LOG.error("Redis is unable to start")
        LOG.error('Check to see if there is an unresponsive instance of redis with "ps -e"')
        LOG.error(redis_out.strip("\n"))
        return False

    redis_out["image_pid"] = redis_out.pop("pid")
    redis_out["parent_pid"] = process.pid
    redis_out["hostname"] = socket.gethostname()
    if not dump_process_file(redis_out, server_config.container.get_pfile_path()):
        LOG.error("Unable to create process file for container.")
        return False

    if get_server_status() != ServerStatus.RUNNING:
        LOG.error("Unable to start merlin server.")
        return False

    LOG.info(f"Server started with PID {str(process.pid)}.")
    LOG.info(f'Merlin server operating on "{redis_out["hostname"]}" and port "{redis_out["port"]}".')
    return True


def start_server() -> bool:  # pylint: disable=R0911
    """
    Starts a Merlin server container.

    This function performs several steps to start the server, including checking
    for an existing non-running server, pulling the server configuration, starting
    the container, verifying the server startup, and applying Redis user and
    configuration settings. It also generates a new `app.yaml` file for the server
    configuration.

    Returns:
        True if the server was successfully started, False otherwise.
    """
    if not check_for_not_running_server():
        return False

    server_config = pull_server_config()
    if not server_config:
        LOG.error('Try to run "merlin server init" again to reinitialize values.')
        return False

    process = start_container(server_config)
    if process is None:
        return False

    if not server_started(process, server_config):
        return False

    redis_users = RedisUsers(server_config.container.get_user_file_path())
    redis_config = RedisConfig(server_config.container.get_config_path())
    redis_users.apply_to_redis(redis_config.get_ip_address(), redis_config.get_port(), redis_config.get_password())

    new_app_yaml = os.path.join(server_config.container.get_config_dir(), "app.yaml")
    app_yaml = AppYaml()
    app_yaml.apply_server_config(server_config=server_config)
    app_yaml.write(new_app_yaml)
    LOG.info(f"New app.yaml written to {new_app_yaml}.")
    LOG.info("Replace app.yaml in ~/.merlin/app.yaml to use merlin server as main configuration.")
    LOG.info("To use for local runs, move app.yaml into the running directory.")

    return True


def stop_server() -> bool:
    """
    Stops a running Merlin server container.

    This function checks the current server status, retrieves the server configuration,
    and attempts to terminate the running server process. If successful, the server
    process is stopped, and the function returns True. Otherwise, it logs errors
    and returns False.

    Returns:
        True if the server was successfully stopped, False otherwise.
    """
    if get_server_status() != ServerStatus.RUNNING:
        LOG.info("There is no instance of merlin server running.")
        LOG.info("Start a merlin server first with 'merlin server start'")
        return False

    server_config = pull_server_config()
    if not server_config:
        LOG.error('Try to run "merlin server init" again to reinitialize values.')
        return False

    pf_data = pull_process_file(server_config.container.get_pfile_path())
    read_pid = pf_data["parent_pid"]

    process = subprocess.run(
        server_config.process.get_status_command().strip("\\").format(pid=read_pid).split(), stdout=subprocess.PIPE
    )
    if process.stdout == b"":
        LOG.error("Unable to get the PID for the current merlin server.")
        return False

    command = server_config.process.get_kill_command().strip("\\").format(pid=read_pid).split()
    if server_config.container_format.get_stop_command() != "kill":
        command = (
            server_config.container_format.get_stop_command()
            .strip("\\")
            .format(name=server_config.container.get_image_name)
            .split()
        )

    LOG.info(f"Attempting to close merlin server PID {str(read_pid)}")

    subprocess.run(command, stdout=subprocess.PIPE)
    time.sleep(1)
    if get_server_status() == ServerStatus.RUNNING:
        LOG.error("Unable to kill process.")
        return False

    LOG.info("Merlin server terminated.")
    return True


def restart_server() -> bool:
    """
    Restarts a running Merlin server instance.

    This function stops the currently running Merlin server and then starts it again.
    If the server is not running, it logs a message and returns False.

    Returns:
        True if the server was successfully restarted, False otherwise.
    """
    if get_server_status() != ServerStatus.RUNNING:
        LOG.info("Merlin server is not currently running.")
        LOG.info("Please start a merlin server instance first with 'merlin server start'")
        return False
    stop_server()
    time.sleep(1)
    start_server()
    return True



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for querying and displaying the status of Merlin workflows.

This module defines two primary CLI command handlers:

- `StatusCommand`: Handles the `status` command, which shows a high-level summary
  of the current state of a Merlin study or workflow. It supports querying status
  based on a Merlin YAML specification file or an existing study workspace, with
  options for output formatting and filtering.

- `DetailedStatusCommand`: Extends `StatusCommand` to provide a more granular,
  task-by-task view of the workflow status via the `detailed-status` command.
  This includes extensive filtering and display customization options.
"""

# pylint: disable=duplicate-code

import logging
from argparse import ArgumentParser, Namespace

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.spec.expansion import get_spec_with_expansion
from merlin.study.status import DetailedStatus, Status
from merlin.study.status_constants import VALID_RETURN_CODES, VALID_STATUS_FILTERS
from merlin.study.status_renderers import status_renderer_factory
from merlin.utils import verify_dirpath, verify_filepath


LOG = logging.getLogger("merlin")


class StatusCommand(CommandEntryPoint):
    """
    Handles `status` CLI command for checking the high-level status of a workflow.

    Methods:
        add_parser: Adds the `status` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `status` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `status` command parser will be added.
        """
        status_cmd: ArgumentParser = subparsers.add_parser(
            "status",
            help="Display a summary of the status of a study.",
        )
        status_cmd.set_defaults(func=self.process_command, detailed=False)
        status_cmd.add_argument(
            "spec_or_workspace", type=str, help="Path to a Merlin YAML spec file or a launched Merlin study"
        )
        status_cmd.add_argument(
            "--cb-help", action="store_true", help="Colorblind help; uses different symbols to represent different statuses"
        )
        status_cmd.add_argument(
            "--dump", type=str, help="Dump the status to a file. Provide the filename (must be .csv or .json).", default=None
        )
        status_cmd.add_argument(
            "--no-prompts",
            action="store_true",
            help="Ignore any prompts provided. This will default to the latest study \
                if you provide a spec file rather than a study workspace.",
        )
        status_cmd.add_argument(
            "--task_server",
            type=str,
            default="celery",
            help="Task server type.\
                                Default: %(default)s",
        )
        status_cmd.add_argument(
            "-o",
            "--output-path",
            action="store",
            type=str,
            default=None,
            help="Specify a location to look for output workspaces. Only used when a spec file is passed as the argument "
            "to 'status'; this will NOT be used if an output workspace is passed as the argument.",
        )

    def process_command(self, args: Namespace):
        """
        CLI command for querying the status of studies.

        This function processes the given command-line arguments to determine the
        status of a study. It constructs either a [`Status`][study.status.Status] object
        or a [`DetailedStatus`][study.status.DetailedStatus] object based on the specified
        command and the arguments provided. The function handles validations for the task
        server input and the output format specified for status dumping.

        Object mapping:
            - `merlin status` -> [`Status`][study.status.Status] object
            - `merlin detailed-status` -> [`DetailedStatus`][study.status.DetailedStatus]
            object

        Args:
            args: Parsed CLI arguments containing user inputs for the status query.

        Raises:
            ValueError:
                - If the task server specified is not supported (only "celery" is valid).
                - If the --dump filename provided does not end with ".csv" or ".json".
        """
        print(banner_small)

        # Ensure task server is valid
        if args.task_server != "celery":
            raise ValueError("Currently the only supported task server is celery.")

        # Make sure dump is valid if provided
        if args.dump and (not args.dump.endswith(".csv") and not args.dump.endswith(".json")):
            raise ValueError("The --dump option takes a filename that must end with .csv or .json")

        # Establish whether the argument provided by the user was a spec file or a study directory
        spec_display = False
        try:
            file_or_ws = verify_filepath(args.spec_or_workspace)
            spec_display = True
        except ValueError:
            try:
                file_or_ws = verify_dirpath(args.spec_or_workspace)
            except ValueError:
                LOG.error(f"The file or directory path {args.spec_or_workspace} does not exist.")
                return

        # If we're loading status based on a spec, load in the spec provided
        if spec_display:
            args.specification = file_or_ws
            args.spec_provided = get_spec_with_expansion(args.specification)

        # Get either a Status object or DetailedStatus object
        if args.detailed:
            status_obj = DetailedStatus(args, spec_display, file_or_ws)
        else:
            status_obj = Status(args, spec_display, file_or_ws)

        # Handle output appropriately
        if args.dump:
            status_obj.dump()
        else:
            status_obj.display()


class DetailedStatusCommand(StatusCommand):
    """
    Handles `detailed-status` CLI command for checking the in-depth status of a workflow.

    Methods:
        add_parser: Adds the `detailed-status` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `detailed-status` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `detailed-status` command parser will be added.
        """
        detailed_status: ArgumentParser = subparsers.add_parser(
            "detailed-status",
            help="Display a task-by-task status of a study.",
        )
        detailed_status.set_defaults(func=self.process_command, detailed=True)
        detailed_status.add_argument(
            "spec_or_workspace", type=str, help="Path to a Merlin YAML spec file or a launched Merlin study"
        )
        detailed_status.add_argument(
            "--dump", type=str, help="Dump the status to a file. Provide the filename (must be .csv or .json).", default=None
        )
        detailed_status.add_argument(
            "--task_server",
            type=str,
            default="celery",
            help="Task server type.\
                                Default: %(default)s",
        )
        detailed_status.add_argument(
            "-o",
            "--output-path",
            action="store",
            type=str,
            default=None,
            help="Specify a location to look for output workspaces. Only used when a spec file is passed as the argument "
            "to 'status'; this will NOT be used if an output workspace is passed as the argument.",
        )
        status_filter_group = detailed_status.add_argument_group("filter options")
        status_filter_group.add_argument(
            "--max-tasks", action="store", type=int, help="Sets a limit on how many tasks can be displayed"
        )
        status_filter_group.add_argument(
            "--return-code",
            action="store",
            nargs="+",
            type=str,
            choices=VALID_RETURN_CODES,
            help="Filter which tasks to display based on their return code",
        )
        status_filter_group.add_argument(
            "--steps",
            nargs="+",
            type=str,
            dest="steps",
            default=["all"],
            help="Filter which tasks to display based on the steps they're associated with",
        )
        status_filter_group.add_argument(
            "--task-queues",
            nargs="+",
            type=str,
            help="Filter which tasks to display based on the task queue they're in",
        )
        status_filter_group.add_argument(
            "--task-status",
            action="store",
            nargs="+",
            type=str,
            choices=VALID_STATUS_FILTERS,
            help="Filter which tasks to display based on their status",
        )
        status_filter_group.add_argument(
            "--workers",
            nargs="+",
            type=str,
            help="Filter which tasks to display based on which workers are processing them",
        )
        status_display_group = detailed_status.add_argument_group("display options")
        status_display_group.add_argument(
            "--disable-pager", action="store_true", help="Turn off the pager functionality when viewing the status"
        )
        status_display_group.add_argument(
            "--disable-theme",
            action="store_true",
            help="Turn off styling for the status layout (If you want styling but it's not working, try modifying "
            "the MANPAGER or PAGER environment variables to be 'less -r'; i.e. export MANPAGER='less -r')",
        )
        status_display_group.add_argument(
            "--layout",
            type=str,
            choices=status_renderer_factory.get_layouts(),
            default="default",
            help="Alternate status layouts [Default: %(default)s]",
        )
        status_display_group.add_argument(
            "--no-prompts",
            action="store_true",
            help="Ignore any prompts provided. This will default to the latest study \
                if you provide a spec file rather than a study workspace.",
        )



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module provides an adapter to the Celery Distributed Task Queue.
"""
import logging
import os
import socket
import subprocess
import time
from contextlib import suppress
from datetime import datetime
from types import SimpleNamespace
from typing import Dict, List, Set, Tuple

from amqp.exceptions import ChannelError
from celery import Celery
from tabulate import tabulate

from merlin.common.dumper import dump_handler
from merlin.config import Config
from merlin.spec.specification import MerlinSpec
from merlin.study.batch import batch_check_parallel, batch_worker_launch
from merlin.study.study import MerlinStudy
from merlin.utils import apply_list_of_regex, check_machines, get_procs, get_yaml_var, is_running


LOG = logging.getLogger(__name__)

# TODO figure out a better way to handle the import of celery app and CONFIG


def run_celery(study: MerlinStudy, run_mode: str = None):
    """
    Run the given [`MerlinStudy`][study.study.MerlinStudy] object with optional
    Celery configuration.

    This function executes the provided [`MerlinStudy`][study.study.MerlinStudy]
    object. If the `run_mode` is set to "local", it configures Celery to run in
    local mode (without utilizing workers). Otherwise, it connects to the Celery
    server to queue tasks.

    Args:
        study (study.study.MerlinStudy): The study object to be executed.
        run_mode: The mode in which to run the study. If set to "local",
            Celery runs locally.
    """
    # Only import celery stuff if we want celery in charge
    # Pylint complains about circular import between merlin.common.tasks -> merlin.router -> merlin.study.celeryadapter
    # For now I think this is still the best way to do this so we'll ignore it
    from merlin.celery import app  # pylint: disable=C0415
    from merlin.common.tasks import queue_merlin_study  # pylint: disable=C0415, R0401

    adapter_config = study.get_adapter_config(override_type="local")

    if run_mode == "local":
        app.conf.task_always_eager = True
        app.conf.task_eager_propogates = True
    else:
        # Check for server
        app.connection().connect()

    # Send the tasks to the server
    queue_merlin_study(study, adapter_config)


def get_running_queues(celery_app_name: str, test_mode: bool = False) -> List[str]:
    """
    Check for running Celery workers and retrieve their associated queues.

    This function inspects currently running processes to identify active
    Celery workers. It extracts queue names from the `-Q` tag in the
    command line of the worker processes. The returned list contains
    only unique Celery queue names. This function must be executed
    on the allocation where the workers are running.

    Note:
        Unlike [`get_active_celery_queues`][study.celeryadapter.get_active_celery_queues],
        this function does _not_ go through the application's server.

    Args:
        celery_app_name: The name of the Celery app (typically "merlin"
            unless in test mode).
        test_mode: If True, the function runs in test mode.

    Returns:
        A unique list of Celery queue names with workers attached to them.
    """
    running_queues = []

    if not is_running(f"{celery_app_name} worker"):
        return running_queues

    proc_name = "celery" if not test_mode else "sh"
    procs = get_procs(proc_name)
    for _, lcmd in procs:
        lcmd = list(filter(None, lcmd))
        cmdline = " ".join(lcmd)
        if "-Q" in cmdline:
            if test_mode:
                echo_cmd = lcmd.pop(2)
                lcmd.extend(echo_cmd.split())
            running_queues.extend(lcmd[lcmd.index("-Q") + 1].split(","))

    running_queues = list(set(running_queues))

    return running_queues


def get_active_celery_queues(app: Celery) -> Tuple[Dict[str, List[str]], List[str]]:
    """
    Retrieve all active queues and their associated workers for a Celery application.

    This function queries the application's server to obtain a comprehensive
    view of active queues and the workers connected to them. It returns a
    dictionary where each key is a queue name and the value is a list of
    workers attached to that queue. Additionally, it provides a list of all
    active workers in the application.

    Note:
        Unlike [`get_running_queues`][study.celeryadapter.get_running_queues],
        this function goes through the application's server.

    Args:
        app: The Celery application instance.

    Returns:
        A tuple containing:\n
            - A dictionary mapping queue names to lists of workers connected to them.
            - A list of all active workers in the application.

    Example:
        ```python
        from merlin.celery import app
        queues, workers = get_active_celery_queues(app)
        queue_names = list(queues)
        workers_on_q0 = queues[queue_names[0]]
        workers_not_on_q0 = [worker for worker in workers if worker not in workers_on_q0]
        ```
    """
    i = app.control.inspect()
    active_workers = i.active_queues()
    if active_workers is None:
        active_workers = {}
    queues = {}
    for worker in active_workers:
        for my_queue in active_workers[worker]:
            try:
                queues[my_queue["name"]].append(worker)
            except KeyError:
                queues[my_queue["name"]] = [worker]
    return queues, [*active_workers]


def get_active_workers(app: Celery) -> Dict[str, List[str]]:
    """
    Retrieve a mapping of active workers to their associated queues for a Celery application.

    This function serves as the inverse of
    [`get_active_celery_queues()`][study.celeryadapter.get_active_celery_queues]. It constructs
    a dictionary where each key is a worker's name and the corresponding value is a
    list of queues that the worker is connected to. This allows for easy identification
    of which queues are being handled by each worker.

    Args:
        app: The Celery application instance.

    Returns:
        A dictionary mapping active worker names to lists of queue names they are
            attached to. If no active workers are found, an empty dictionary is returned.
    """
    # Get the information we need from celery
    i = app.control.inspect()
    active_workers = i.active_queues()
    if active_workers is None:
        active_workers = {}

    # Build the mapping dictionary
    worker_queue_map = {}
    for worker, queues in active_workers.items():
        for queue in queues:
            if worker in worker_queue_map:
                worker_queue_map[worker].append(queue["name"])
            else:
                worker_queue_map[worker] = [queue["name"]]

    return worker_queue_map


def celerize_queues(queues: List[str], config: SimpleNamespace = None):
    """
    Prepend a queue tag to each queue in the provided list to conform to Celery's
    queue naming requirements.

    This function modifies the input list of queues by adding a specified queue tag
    from the configuration. If no configuration is provided, it defaults to using
    the global configuration settings.

    Args:
        queues: A list of queue names that need the queue tag prepended.
        config: A SimpleNamespace of configuration settings. If not provided, the
            function will use the default configuration.
    """
    if config is None:
        from merlin.config.configfile import CONFIG as config  # pylint: disable=C0415

    for i, queue in enumerate(queues):
        queues[i] = f"{config.celery.queue_tag}{queue}"


def _build_output_table(worker_list: List[str], output_table: List[Tuple[str, str]]):
    """
    Construct an output table for displaying the status of workers and their associated queues.

    This helper function populates the provided output table with entries for each worker
    in the given worker list. It retrieves the mapping of active workers to their queues
    and formats the data accordingly.

    Args:
        worker_list: A list of worker names to be included in the output table.
        output_table: A list of tuples where each entry will be of the form
            (worker name, associated queues).
    """
    from merlin.celery import app  # pylint: disable=C0415

    # Get a mapping between workers and the queues they're watching
    worker_queue_map = get_active_workers(app)

    # Loop through the list of workers and add an entry in the table
    # of the form (worker name, queues attached to this worker)
    for worker in worker_list:
        if "celery@" not in worker:
            worker = f"celery@{worker}"
        output_table.append((worker, ", ".join(worker_queue_map[worker])))


def query_celery_workers(spec_worker_names: List[str], queues: List[str], workers_regex: List[str]):
    """
    Query and filter existing Celery workers based on specified criteria,
    and print a table of the workers along with their associated queues.

    This function retrieves the list of active Celery workers and filters them
    according to the provided specifications, including worker names from a
    spec file, specific queues, and regular expressions for worker names.
    It then constructs and displays a table of the matching workers and their
    associated queues.

    Args:
        spec_worker_names: A list of worker names defined in a spec file
            to filter the workers.
        queues: A list of queues to filter the workers by.
        workers_regex: A list of regular expressions to filter the worker names.
    """
    from merlin.celery import app  # pylint: disable=C0415

    # Ping all workers and grab which ones are running
    workers = get_workers_from_app()
    if not workers:
        LOG.warning("No workers found!")
        return

    # Remove prepended celery tag while we filter
    workers = [worker.replace("celery@", "") for worker in workers]
    workers_to_query = []

    # --queues flag
    if queues:
        # Get a mapping between queues and the workers watching them
        queue_worker_map, _ = get_active_celery_queues(app)
        # Remove duplicates and prepend the celery queue tag to all queues
        queues = list(set(queues))
        celerize_queues(queues)
        # Add the workers associated to each queue to the list of workers we're
        # going to query
        for queue in queues:
            try:
                workers_to_query.extend(queue_worker_map[queue])
            except KeyError:
                LOG.warning(f"No workers connected to {queue}.")

    # --spec flag
    if spec_worker_names:
        apply_list_of_regex(spec_worker_names, workers, workers_to_query)

    # --workers flag
    if workers_regex:
        apply_list_of_regex(workers_regex, workers, workers_to_query)

    # Remove any potential duplicates
    workers = set(workers)
    workers_to_query = set(workers_to_query)

    # If there were filters and nothing was found then we can't display a table
    if (queues or spec_worker_names or workers_regex) and not workers_to_query:
        LOG.warning("No workers found that match your filters.")
        return

    # Build the output table based on our filters
    table = []
    if workers_to_query:
        _build_output_table(workers_to_query, table)
    else:
        _build_output_table(workers, table)

    # Display the output table
    LOG.info("Found these connected workers:")
    print(tabulate(table, headers=["Workers", "Queues"]))
    print()


def build_csv_queue_info(query_return: List[Tuple[str, int, int]], date: str) -> Dict[str, List]:
    """
    Construct a dictionary containing queue information and column labels
    for writing to a CSV file.

    This function processes the output from the [`query_queues`][router.query_queues]
    function and organizes the data into a format suitable for CSV export. It includes
    a timestamp to indicate when the status was recorded.

    Args:
        query_return: The output from the [`query_queues`][router.query_queues] function,
            containing queue names and their associated statistics.
        date: A timestamp indicating when the queue status was recorded.

    Returns:
        A dictionary where keys are column labels and values are lists containing the
            corresponding queue information, formatted for CSV output.
    """
    # Build the list of labels if necessary
    csv_to_dump = {"time": [date]}
    for queue_name, queue_stats in query_return.items():
        csv_to_dump[f"{queue_name}:tasks"] = [str(queue_stats["jobs"])]
        csv_to_dump[f"{queue_name}:consumers"] = [str(queue_stats["consumers"])]

    return csv_to_dump


def build_json_queue_info(query_return: List[Tuple[str, int, int]], date: str) -> Dict:
    """
    Construct a dictionary containing queue information for JSON export.

    This function processes the output from the [`query_queues`][router.query_queues]
    function and organizes the data into a structured format suitable for JSON
    serialization. It includes a timestamp to indicate when the queue status was
    recorded.

    Args:
        query_return: The output from the [`query_queues`][router.query_queues]
            function, containing queue names and their associated statistics.
        date: A timestamp indicating when the queue status was recorded.

    Returns:
        A dictionary structured for JSON output, where the keys are timestamps
            and the values are dictionaries containing queue names and their
            corresponding statistics (tasks and consumers).
    """
    # Get the datetime so we can track different entries and initalize a new json entry
    json_to_dump = {date: {}}

    # Add info for each queue (name)
    for queue_name, queue_stats in query_return.items():
        json_to_dump[date][queue_name] = {"tasks": queue_stats["jobs"], "consumers": queue_stats["consumers"]}

    return json_to_dump


def dump_celery_queue_info(query_return: List[Tuple[str, int, int]], dump_file: str):
    """
    Format and dump Celery queue information to a specified file.

    This function processes the output from the `query_queues` function, formats
    the data according to the file type (CSV or JSON), and adds a timestamp
    to the information before writing it to the specified file.

    Args:
        query_return: The output from the [`query_queues`][router.query_queues]
            function, containing queue names and their associated statistics.
        dump_file: The filepath of the file where the queue information
            will be written. The file extension determines the format (CSV or JSON).
    """
    # Get a timestamp for this dump
    date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    # Handle different file types
    if dump_file.endswith(".csv"):
        # Build the lists of information/labels we'll need
        dump_info = build_csv_queue_info(query_return, date)
    elif dump_file.endswith(".json"):
        # Build the dict of info to dump to the json file
        dump_info = build_json_queue_info(query_return, date)

    # Dump the information
    dump_handler(dump_file, dump_info)


def _get_specific_queues(queues: Set[str], specific_queues: List[str], spec: MerlinSpec, verbose: bool = True) -> Set[str]:
    """
    Retrieve a set of specific queues requested by the user, filtering out those that do not exist.

    This function checks a provided list of specific queues against a set of existing queues
    (from a [`MerlinSpec`][spec.specification.MerlinSpec] object) and returns a set of queues
    that are found. If a queue is not found in the existing set, it will be excluded from the
    results. The function also logs messages based on the verbosity setting.

    Args:
        queues: A set of existing queues, which may be empty or populated from the `spec`
            object.
        specific_queues: A list of specific queue names to search for.
        spec (spec.specification.MerlinSpec): A [`MerlinSpec`][spec.specification.MerlinSpec]
            object that may provide context for the search. Can be None.
        verbose: If True, log messages will be displayed.

    Returns:
        A set containing the specific queues that were found in the existing queues.
    """
    if verbose:
        LOG.info(f"Filtering queues to query by these specific queues: {specific_queues}")

    # Add the queue tag from celery if necessary
    celerize_queues(specific_queues)
    # Remove any potential duplicates and create a new set to store the queues we'll want to check
    specific_queues = set(specific_queues)
    specific_queues_to_check = set()

    for specific_queue in specific_queues:
        # If the user provided the --spec flag too then we'll need to check that this
        # specific queue exists in the spec that they provided
        add_specific_queue = True
        if spec and specific_queue not in queues:
            if verbose:
                LOG.warning(
                    f"Either couldn't find {specific_queue} in the existing queues for the spec file provided or "
                    "this queue doesn't go with the steps provided with the --steps option. Ignoring this queue."
                )
            add_specific_queue = False

        # Add the full queue name to the set of queues we'll check
        if add_specific_queue:
            specific_queues_to_check.add(specific_queue)

    return specific_queues_to_check


def build_set_of_queues(
    spec: MerlinSpec,
    steps: List[str],
    specific_queues: List[str],
    verbose: bool = True,
    app: Celery = None,
) -> Set[str]:
    """
    Construct a set of queues to query based on the provided parameters.

    This function builds a set of queues by querying a [`MerlinSpec`][spec.specification.MerlinSpec]
    object for queues associated with specified steps and/or filtering for specific queue names.
    If no spec or specific queues are provided, it defaults to querying active queues from the Celery
    application.

    Args:
        spec (spec.specification.MerlinSpec): A [`MerlinSpec`][spec.specification.MerlinSpec]
            object that defines the context for the query. Can be None.
        steps: A list of step names to query. If empty, all steps are considered.
        specific_queues: A list of specific queue names to filter. Can be None.
        verbose: If True, log statements will be output. Defaults to True.
        app: A Celery application instance. If None, it will be imported.

    Returns:
        A set of queue names to investigate based on the provided parameters.
    """
    if app is None:
        from merlin.celery import app  # pylint: disable=C0415

    queues = set()
    # If the user provided a spec file, get the queues from that spec
    if spec:
        if verbose:
            LOG.info(f"Querying queues for steps = {steps}")
        queues = set(spec.get_queue_list(steps))

    # If the user provided specific queues, search for those
    if specific_queues:
        queues = _get_specific_queues(queues, specific_queues, spec, verbose=verbose)

    # Default behavior with no options provided; display active queues
    if not spec and not specific_queues:
        if verbose:
            LOG.info("Querying active queues")
        existing_queues, _ = get_active_celery_queues(app)

        # Check if there's any active queues currently
        if len(existing_queues) == 0:
            if verbose:
                LOG.warning("No active queues found. Are your workers running yet?")
            return set()

        # Set the queues we're going to check to be all existing queues by default
        queues = set(existing_queues.keys())

    return queues


def query_celery_queues(queues: List[str], app: Celery = None, config: Config = None) -> Dict[str, Dict[str, int]]:
    """
    Retrieve information about the number of jobs and consumers for specified Celery queues.

    This function constructs a dictionary containing details about the number of jobs
    and consumers associated with each queue provided in the input list. It connects
    to the Celery application to gather this information, handling both Redis and
    RabbitMQ brokers.

    Notes:
        - If the specified queue does not exist or has no jobs, it will be handled gracefully.
        - For Redis brokers, the function counts consumers by inspecting active queues
          since Redis does not track consumers like RabbitMQ does.

    Args:
        queues: A list of queue names for which to gather information.
        app: The Celery application instance. Defaults to None, which triggers an import
            for testing purposes.
        config (config.Config): A configuration object containing broker details.
            Defaults to None, which also triggers an import for testing.

    Returns:
        A dictionary where each key is a queue name and the value is another dictionary
            containing:\n
            - `jobs`: The number of jobs in the queue.
            - `consumers`: The number of consumers attached to the queue.
    """
    if app is None:
        from merlin.celery import app  # pylint: disable=C0415
    if config is None:
        from merlin.config.configfile import CONFIG as config  # pylint: disable=C0415

    # Initialize the dictionary with the info we want about our queues
    queue_info = {queue: {"consumers": 0, "jobs": 0} for queue in queues}

    # Open a connection via our Celery app
    with app.connection() as conn:
        # Open a channel inside our connection
        with conn.channel() as channel:
            # Loop through all the queues we're searching for
            for queue in queues:
                try:
                    # Count the number of jobs and consumers for each queue
                    _, queue_info[queue]["jobs"], queue_info[queue]["consumers"] = channel.queue_declare(
                        queue=queue, passive=True
                    )
                # Redis likes to throw this error when a queue we're looking for has no jobs
                except ChannelError:
                    pass

    # Redis doesn't keep track of consumers attached to queues like rabbit does
    # so we have to count this ourselves here
    if config.broker.name in ("rediss", "redis"):
        # Get a dict of active queues by querying the celery app
        active_queues = app.control.inspect().active_queues()
        if active_queues is not None:
            # Loop through each active queue that was found
            for active_queue_list in active_queues.values():
                # Loop through each queue that each worker is watching
                for active_queue in active_queue_list:
                    # If this is a queue we're looking for, increment the consumer count
                    if active_queue["name"] in queues:
                        queue_info[active_queue["name"]]["consumers"] += 1

    return queue_info


def get_workers_from_app() -> List[str]:
    """
    Retrieve a list of all workers connected to the Celery application.

    This function uses the Celery control interface to inspect the current state
    of the application and returns a list of workers that are currently connected.
    If no workers are found, an empty list is returned.

    Returns:
        A list of worker names that are currently connected to the Celery application.
            If no workers are connected, an empty list is returned.
    """
    from merlin.celery import app  # pylint: disable=C0415

    i = app.control.inspect()
    workers = i.ping()
    if workers is None:
        return []
    return [*workers]


def check_celery_workers_processing(queues_in_spec: List[str], app: Celery) -> bool:
    """
    Check if any Celery workers are currently processing tasks from specified queues.

    This function queries the Celery application to determine if there are any active
    tasks being processed by workers for the given list of queues. It returns a boolean
    indicating whether any tasks are currently active.

    Args:
        queues_in_spec: A list of queue names to check for active tasks.
        app: The Celery application instance used for querying.

    Returns:
        True if any workers are processing tasks in the specified queues; False
            otherwise.
    """
    # Query celery for active tasks
    active_tasks = app.control.inspect().active()

    # Search for the queues we provided if necessary
    if active_tasks is not None:
        for tasks in active_tasks.values():
            for task in tasks:
                if task["delivery_info"]["routing_key"] in queues_in_spec:
                    return True

    return False


def _get_workers_to_start(spec: MerlinSpec, steps: List[str]) -> Set[str]:
    """
    Determine the set of workers to start based on the specified steps.

    This helper function retrieves a mapping of steps to their corresponding workers
    from a [`MerlinSpec`][spec.specification.MerlinSpec] object and returns a unique
    set of workers that should be started for the provided list of steps. If a step
    is not found in the mapping, a warning is logged.

    Args:
        spec (spec.specification.MerlinSpec): An instance of the
            [`MerlinSpec`][spec.specification.MerlinSpec] class that contains the
            mapping of steps to workers.
        steps: A list of steps for which workers need to be started.

    Returns:
        A set of unique workers to be started based on the specified steps.
    """
    workers_to_start = []
    step_worker_map = spec.get_step_worker_map()
    for step in steps:
        try:
            workers_to_start.extend(step_worker_map[step])
        except KeyError:
            LOG.warning(f"Cannot start workers for step: {step}. This step was not found.")

    workers_to_start = set(workers_to_start)
    LOG.debug(f"workers_to_start: {workers_to_start}")

    return workers_to_start


def _create_kwargs(spec: MerlinSpec) -> Tuple[Dict[str, str], Dict]:
    """
    Construct the keyword arguments for launching a worker process.

    This helper function creates a dictionary of keyword arguments that will be
    passed to `subprocess.Popen` when launching a worker. It retrieves the
    environment variables defined in a [`MerlinSpec`][spec.specification.MerlinSpec]
    object and updates the shell environment accordingly.

    Args:
        spec (spec.specification.MerlinSpec): An instance of the MerlinSpec class
            that contains environment specifications.

    Returns:
        A tuple containing:
            - A dictionary of keyword arguments for `subprocess.Popen`, including
              the updated environment.
            - A dictionary of variables defined in the spec, or None if no variables
              were defined.
    """
    # Get the environment from the spec and the shell
    spec_env = spec.environment
    shell_env = os.environ.copy()
    yaml_vars = None

    # If the environment from the spec has anything in it,
    # read in the variables and save them to the shell environment
    if spec_env:
        yaml_vars = get_yaml_var(spec_env, "variables", {})
        for var_name, var_val in yaml_vars.items():
            shell_env[str(var_name)] = str(var_val)
            # For expandvars
            os.environ[str(var_name)] = str(var_val)

    # Create the kwargs dict
    kwargs = {"env": shell_env, "shell": True, "universal_newlines": True}
    return kwargs, yaml_vars


def _get_steps_to_start(wsteps: List[str], steps: List[str], steps_provided: bool) -> List[str]:
    """
    Identify the steps for which workers should be started.

    This function determines which steps to initiate based on the steps
    associated with a worker and the user-provided steps. If specific steps
    are provided by the user, only those steps that match the worker's steps
    will be included. If no specific steps are provided, all worker-associated
    steps will be returned.

    Args:
        wsteps: A list of steps that are associated with a worker.
        steps: A list of steps specified by the user to start workers for.
        steps_provided: A boolean indicating whether the user provided
            specific steps to start.

    Returns:
        A list of steps for which workers should be started.
    """
    steps_to_start = []
    if steps_provided:
        for wstep in wsteps:
            if wstep in steps:
                steps_to_start.append(wstep)
    else:
        steps_to_start.extend(wsteps)

    return steps_to_start


def start_celery_workers(
    spec: MerlinSpec, steps: List[str], celery_args: str, disable_logs: bool, just_return_command: bool
) -> str:  # pylint: disable=R0914,R0915
    """
    Start Celery workers based on the provided specifications and steps.

    This function initializes and starts Celery workers for the specified steps
    in the given [`MerlinSpec`][spec.specification.MerlinSpec]. It constructs
    the necessary command-line arguments and handles the launching of subprocesses
    for each worker. If the `just_return_command` flag is set to `True`, it will
    return the command(s) to start the workers without actually launching them.

    Args:
        spec (spec.specification.MerlinSpec): A [`MerlinSpec`][spec.specification.MerlinSpec]
            object representing the study configuration.
        steps: A list of steps for which to start workers.
        celery_args: A string of additional arguments to pass to the Celery workers.
        disable_logs: A flag to disable logging for the Celery workers.
        just_return_command: If `True`, returns the launch command(s) without starting the workers.

    Returns:
        A string containing all the worker launch commands.

    Side Effects:
        - Starts subprocesses for each worker that is launched, so long as `just_return_command`
          is not True.

    Example:
        Below is an example configuration for Merlin workers:

        ```yaml
        merlin:
          resources:
            task_server: celery
            overlap: False
            workers:
                simworkers:
                    args: -O fair --prefetch-multiplier 1 -E -l info --concurrency 4
                    steps: [run, data]
                    nodes: 1
                    machine: [hostA, hostB]
        ```
    """
    if not just_return_command:
        LOG.info("Starting workers")

    overlap = spec.merlin["resources"]["overlap"]
    workers = spec.merlin["resources"]["workers"]

    # Build kwargs dict for subprocess.Popen to use when we launch the worker
    kwargs, yenv = _create_kwargs(spec)

    worker_list = []
    local_queues = []

    # Get the workers we need to start if we're only starting certain steps
    steps_provided = False if "all" in steps else True  # pylint: disable=R1719
    if steps_provided:
        workers_to_start = _get_workers_to_start(spec, steps)

    for worker_name, worker_val in workers.items():
        # Only triggered if --steps flag provided
        if steps_provided and worker_name not in workers_to_start:
            continue

        skip_loop_step: bool = examine_and_log_machines(worker_val, yenv)
        if skip_loop_step:
            continue

        worker_args = get_yaml_var(worker_val, "args", celery_args)
        with suppress(KeyError):
            if worker_val["args"] is None:
                worker_args = ""

        worker_nodes = get_yaml_var(worker_val, "nodes", None)
        worker_batch = get_yaml_var(worker_val, "batch", None)

        # Get the correct steps to start workers for
        wsteps = get_yaml_var(worker_val, "steps", steps)
        steps_to_start = _get_steps_to_start(wsteps, steps, steps_provided)
        queues = spec.make_queue_string(steps_to_start)

        # Check for missing arguments
        worker_args = verify_args(spec, worker_args, worker_name, overlap, disable_logs=disable_logs)

        # Add a per worker log file (debug)
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("Redirecting worker output to individual log files")
            worker_args += " --logfile %p.%i"

        # Get the celery command & add it to the batch launch command
        celery_com = get_celery_cmd(queues, worker_args=worker_args, just_return_command=True)
        celery_cmd = os.path.expandvars(celery_com)
        worker_cmd = batch_worker_launch(spec, celery_cmd, nodes=worker_nodes, batch=worker_batch)
        worker_cmd = os.path.expandvars(worker_cmd)

        LOG.debug(f"worker cmd={worker_cmd}")

        if just_return_command:
            worker_list = ""
            print(worker_cmd)
            continue

        # Get the running queues
        running_queues = []
        running_queues.extend(local_queues)
        queues = queues.split(",")
        if not overlap:
            running_queues.extend(get_running_queues("merlin"))
            # Cache the queues from this worker to use to test
            # for existing queues in any subsequent workers.
            # If overlap is True, then do not check the local queues.
            # This will allow multiple workers to pull from the same
            # queue.
            local_queues.extend(queues)

        # Search for already existing queues and log a warning if we try to start one that already exists
        found = []
        for q in queues:  # pylint: disable=C0103
            if q in running_queues:
                found.append(q)
        if found:
            LOG.warning(
                f"A celery worker named '{worker_name}' is already configured/running for queue(s) = {' '.join(found)}"
            )
            continue

        # Start the worker
        launch_celery_worker(worker_cmd, worker_list, kwargs)

    # Return a string with the worker commands for logging
    return str(worker_list)


def examine_and_log_machines(worker_val: Dict, yenv: Dict[str, str]) -> bool:
    """
    Determine if a worker should be skipped based on machine availability and log any errors.

    This function checks the specified machines for a worker and determines
    whether the worker can be started. If the machines are not available,
    it logs an error message regarding the output path for the Celery worker.
    If the environment variables (`yenv`) are not provided or do not specify
    an output path, a warning is logged.

    Args:
        worker_val: A dictionary containing worker configuration, including
            the list of machines associated with the worker.
        yenv: A dictionary of environment variables that may include the
            output path for logging.

    Returns:
        Returns `True` if the worker should be skipped (i.e., machines are
            unavailable), otherwise returns `False`.
    """
    worker_machines = get_yaml_var(worker_val, "machines", None)
    if worker_machines:
        LOG.debug(f"check machines = {check_machines(worker_machines)}")
        if not check_machines(worker_machines):
            return True

        if yenv:
            output_path = get_yaml_var(yenv, "OUTPUT_PATH", None)
            if output_path and not os.path.exists(output_path):
                hostname = socket.gethostname()
                LOG.error(f"The output path, {output_path}, is not accessible on this host, {hostname}")
        else:
            LOG.warning(
                "The env:variables section does not have an OUTPUT_PATH specified, multi-machine checks cannot be performed."
            )
        return False
    return False


def verify_args(spec: MerlinSpec, worker_args: str, worker_name: str, overlap: bool, disable_logs: bool = False) -> str:
    """
    Validate and enhance the arguments passed to a Celery worker for completeness.

    This function checks the provided worker arguments to ensure that they include
    recommended settings for running parallel tasks. It adds default values for
    concurrency, prefetch multiplier, and logging level if they are not specified.
    Additionally, it generates a unique worker name based on the current time if
    the `-n` argument is not provided.

    Args:
        spec (spec.specification.MerlinSpec): A [`MerlinSpec`][spec.specification.MerlinSpec]
            object containing the study configuration.
        worker_args: A string of arguments passed to the worker that may need validation.
        worker_name: The name of the worker, used for generating a unique worker identifier.
        overlap: A flag indicating whether multiple workers can overlap in their queue processing.
        disable_logs: A flag to disable logging configuration for the worker.

    Returns:
        The validated and potentially modified worker arguments string.
    """
    parallel = batch_check_parallel(spec)
    if parallel:
        if "--concurrency" not in worker_args:
            LOG.warning("The worker arg --concurrency [1-4] is recommended when running parallel tasks")
        if "--prefetch-multiplier" not in worker_args:
            LOG.warning("The worker arg --prefetch-multiplier 1 is recommended when running parallel tasks")
        if "fair" not in worker_args:
            LOG.warning("The worker arg -O fair is recommended when running parallel tasks")

    if "-n" not in worker_args:
        nhash = ""
        if overlap:
            nhash = time.strftime("%Y%m%d-%H%M%S")
        # TODO: Once flux fixes their bug, change this back to %h
        # %h in Celery is short for hostname including domain name
        worker_args += f" -n {worker_name}{nhash}.%%h"

    if not disable_logs and "-l" not in worker_args:
        worker_args += f" -l {logging.getLevelName(LOG.getEffectiveLevel())}"

    return worker_args


def launch_celery_worker(worker_cmd: str, worker_list: List[str], kwargs: Dict):
    """
    Launch a Celery worker using the specified command and parameters.

    This function executes the provided Celery command to start a worker as a
    subprocess. It appends the command to the given list of worker commands
    for tracking purposes. If the worker fails to start, an error is logged.

    Args:
        worker_cmd: The command string used to launch the Celery worker.
        worker_list: A list that will be updated to include the launched
            worker command for tracking active workers.
        kwargs: A dictionary of additional keyword arguments to pass to
            `subprocess.Popen`, allowing for customization of the subprocess
            behavior.

    Raises:
        Exception: If the worker fails to start, an error is logged, and the
            exception is re-raised.

    Side Effects:
        - Launches a Celery worker process in the background.
        - Modifies the `worker_list` by appending the launched worker command.
    """
    try:
        subprocess.Popen(worker_cmd, **kwargs)  # pylint: disable=R1732
        worker_list.append(worker_cmd)
    except Exception as e:  # pylint: disable=C0103
        LOG.error(f"Cannot start celery workers, {e}")
        raise


def get_celery_cmd(queue_names: str, worker_args: str = "", just_return_command: bool = False) -> str:
    """
    Construct the command to launch Celery workers for the specified queues.

    This function generates a command string that can be used to start Celery
    workers associated with the provided queue names. It allows for optional
    worker arguments to be included and can return the command without executing it.

    Args:
        queue_names: A comma-separated string of the queue name(s) to which the worker
            will be associated.
        worker_args: Additional command-line arguments for the Celery worker.
        just_return_command: If True, the function will return the constructed command
            without executing it.

    Returns:
        The constructed command string for launching the Celery worker. If
            `just_return_command` is True, returns the command; otherwise, returns an
            empty string.
    """
    worker_command = " ".join(["celery -A merlin worker", worker_args, "-Q", queue_names])
    if just_return_command:
        return worker_command
    # If we get down here, this only runs celery locally the user would need to
    # add all of the flux config themselves.
    return ""


def purge_celery_tasks(queues: str, force: bool) -> int:
    """
    Purge Celery tasks from the specified queues.

    This function constructs and executes a command to purge tasks from the
    specified Celery queues. If the `force` parameter is set to True, the
    purge operation will be executed without prompting for confirmation.

    Args:
        queues: A comma-separated string of the queue name(s) from which
            tasks should be purged.
        force: If True, the purge operation will be executed without asking
            for user confirmation.

    Returns:
        The return code from the subprocess execution. A return code of
            0 indicates success, while any non-zero value indicates an error
            occurred during the purge operation.
    """
    # This version will purge all queues.
    # from merlin.celery import app
    # app.control.purge()
    force_com = ""
    if force:
        force_com = " -f "
    purge_command = " ".join(["celery -A merlin purge", force_com, "-Q", queues])
    LOG.debug(purge_command)
    return subprocess.run(purge_command, shell=True).returncode


def stop_celery_workers(
    queues: List[str] = None, spec_worker_names: List[str] = None, worker_regex: List[str] = None
):  # pylint: disable=R0912
    """
    Send a stop command to Celery workers.

    This function sends a shutdown command to Celery workers associated with
    specified queues. By default, it stops all connected workers, but it can
    be configured to target specific workers based on queue names or regular
    expression patterns.

    Args:
        queues: A list of queue names to which the stop command will be sent.
            If None, all connected workers across all queues will be stopped.
        spec_worker_names: A list of specific worker names to stop, in addition
            to those matching the `worker_regex`.
        worker_regex: A regular expression string used to match worker names.
            If None, no regex filtering will be applied.

    Side Effects:
        - Broadcasts a shutdown signal to Celery workers

    Example:
        ```python
        stop_celery_workers(queues=['hello'], worker_regex='celery@*my_machine*')
        stop_celery_workers()
        ```
    """
    from merlin.celery import app  # pylint: disable=C0415

    LOG.debug(f"Sending stop to queues: {queues}, worker_regex: {worker_regex}, spec_worker_names: {spec_worker_names}")
    active_queues, _ = get_active_celery_queues(app)

    # If not specified, get all the queues
    if queues is None:
        queues = [*active_queues]
    # Celery adds the queue tag in front of each queue so we add that here
    else:
        celerize_queues(queues)

    # Find the set of all workers attached to all of those queues
    all_workers = set()
    for queue in queues:
        try:
            all_workers.update(active_queues[queue])
            LOG.debug(f"Workers attached to queue {queue}: {active_queues[queue]}")
        except KeyError:
            LOG.warning(f"No workers are connected to queue {queue}")

    all_workers = list(all_workers)

    LOG.debug(f"Pre-filter worker stop list: {all_workers}")

    # Stop workers with no flags
    if (spec_worker_names is None or len(spec_worker_names) == 0) and worker_regex is None:
        workers_to_stop = list(all_workers)
    # Flag handling
    else:
        workers_to_stop = []
        # --spec flag
        if (spec_worker_names is not None) and len(spec_worker_names) > 0:
            apply_list_of_regex(spec_worker_names, all_workers, workers_to_stop)
        # --workers flag
        if worker_regex is not None:
            LOG.debug(f"Searching for workers to stop based on the following regex's: {worker_regex}")
            apply_list_of_regex(worker_regex, all_workers, workers_to_stop)

    # Remove duplicates
    workers_to_stop = list(set(workers_to_stop))
    LOG.debug(f"Post-filter worker stop list: {workers_to_stop}")

    if workers_to_stop:
        LOG.info(f"Sending stop to these workers: {workers_to_stop}")
        # Send the shutdown signal
        app.control.broadcast("shutdown", destination=workers_to_stop)
    else:
        LOG.warning("No workers found to stop")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Redis backend implementation for the Merlin application.

This module provides a concrete implementation of the `ResultsBackend` interface using Redis
as the underlying database. It defines the `RedisBackend` class, which manages interactions
with Redis-specific store classes for different data models, including CRUD operations and
database flushing.
"""

import logging

from redis import Redis

from merlin.backends.redis.redis_stores import (
    RedisLogicalWorkerStore,
    RedisPhysicalWorkerStore,
    RedisRunStore,
    RedisStudyStore,
)
from merlin.backends.results_backend import ResultsBackend


LOG = logging.getLogger("merlin")


# TODO might be able to make ResultsBackend classes replace the config/results_backend.py file
# - would help get a more OOP approach going within Merlin's codebase
# - instead of calling get_connection_string that logic could be handled in the base class?
class RedisBackend(ResultsBackend):
    """
    A Redis-based implementation of the `ResultsBackend` interface for interacting with a Redis database.

    Attributes:
        backend_name (str): The name of the backend (e.g., "redis").
        client (Redis): The Redis client used for database operations.

    Methods:
        get_version:
            Query Redis for the current version.

        get_connection_string:
            Retrieve the connection string used to connect to Redis.

        flush_database:
            Remove every entry in the Redis database.

        save:
            Save a `BaseDataModel` entity to the Redis database.

        retrieve:
            Retrieve an entity from the appropriate store based on the given query identifier and store type.

        retrieve_all:
            Retrieve all entities from the specified store.

        delete:
            Delete an entity from the specified store.
    """

    def __init__(self, backend_name: str):
        """
        Initialize the `RedisBackend` instance, setting up the Redis client connection and store mappings.

        Args:
            backend_name (str): The name of the backend (e.g., "redis").
        """
        from merlin.config.configfile import CONFIG  # pylint: disable=import-outside-toplevel
        from merlin.config.results_backend import get_connection_string  # pylint: disable=import-outside-toplevel

        # Get the Redis client connection
        redis_config = {"url": get_connection_string(), "decode_responses": True}
        if CONFIG.results_backend.name == "rediss":
            redis_config.update({"ssl_cert_reqs": getattr(CONFIG.results_backend, "cert_reqs", "required")})
        self.client: Redis = Redis.from_url(**redis_config)

        # Create instances of each store in our database
        stores = {
            "study": RedisStudyStore(self.client),
            "run": RedisRunStore(self.client),
            "logical_worker": RedisLogicalWorkerStore(self.client),
            "physical_worker": RedisPhysicalWorkerStore(self.client),
        }

        super().__init__(backend_name, stores)

    def get_version(self) -> str:
        """
        Query the Redis backend for the current version.

        Returns:
            A string representing the current version of Redis.
        """
        client_info = self.client.info()
        return client_info.get("redis_version", "N/A")

    def flush_database(self):
        """
        Remove everything stored in Redis.
        """
        self.client.flushdb()



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module defines all the possible keys for each block in a Merlin specification file.

Merlin specification files are used to configure and manage studies, workflows, and tasks
in Merlin. Each block in the spec file corresponds to a specific aspect of the workflow,
such as batch settings, environment variables, study steps, parameters, and resources.
This module provides sets of valid keys for each block to ensure consistency and validation.

This module serves as a reference for the structure and valid keys of Merlin specification files.
"""

DESCRIPTION = {"description", "name"}

BATCH = {
    "type",
    "bank",
    "queue",
    "dry_run",
    "shell",
    "flux_path",
    "flux_start_opts",
    "flux_exec",
    "flux_exec_workers",
    "launch_pre",
    "launch_args",
    "worker_launch",
    "nodes",
    "walltime",
}

ENV = {"variables", "labels", "sources", "dependencies"}

STUDY_STEP = {"name", "description", "run"}

STUDY_STEP_RUN = {
    "cmd",
    "restart",
    "task_queue",
    "shell",
    "max_retries",
    "depends",
    "nodes",
    "procs",
    "cores per task",
    "gpus per task",
    "batch",
    "slurm",
    "lsf",
    "bind",
    "num resource set",
    "launch_distribution",
    "exit_on_error",
    "flux",
    "walltime",
}

PARAMETER = {"values", "label"}

MERLIN_RESOURCES = {"task_server", "overlap", "workers"}

MERLIN = {"resources", "samples"}

WORKER = {"steps", "nodes", "batch", "args", "machines"}

SAMPLES = {"generate", "level_max_dirs", "file", "column_labels"}



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module provides the `MerlinConfigManager` class, which is responsible for managing
Merlin configuration files. It allows users to initialize, create, and update configuration
files for the Merlin application, including broker and results backend settings.
"""

import logging
import os
import shutil
from argparse import Namespace
from importlib import resources
from typing import Dict, List, Union

import yaml

from merlin.config.config_filepaths import CONFIG_PATH_FILE
from merlin.exceptions import MerlinInvalidTaskServerError


LOG = logging.getLogger(__name__)


class MerlinConfigManager:
    """
    A class to manage the configuration of the Merlin application.

    This class provides functionality to initialize, create, and update configuration files
    for the Merlin application. It supports updating broker and results backend settings
    for Redis and RabbitMQ, and allows users to switch between multiple configurations
    using a `config_path.txt` file.

    Attributes:
        args (Namespace): Parsed command-line arguments.
        output_dir (str): The directory where configuration files are stored.
        config_file (str): The path to the current configuration file.

    Methods:
        save_config_path:
            Saves the path to the configuration file in `config_path.txt`.

        create_template_config:
            Creates a template configuration file based on user input.

        update_broker:
            Updates the broker section of the configuration file.

        update_backend:
            Updates the results backend section of the configuration file.

        update_redis_config:
            Updates Redis-specific configuration fields for broker or backend.

        update_rabbitmq_config:
            Updates RabbitMQ-specific configuration fields for broker.

        update_config:
            Generic method to update configuration fields based on user input.
    """

    def __init__(self, args: Namespace):
        """
        Initialize the configuration manager.

        Args:
            args: Parsed command-line arguments.
        """
        self.args = args
        self.config_file = getattr(args, "config_file", None)
        if not self.config_file:  # This should never be reached because of argparse defaults
            raise ValueError("No config file given to MerlinConfigManager.")

        self.config_file = os.path.abspath(self.config_file)
        LOG.debug(f"MerlinConfigManager successfully initialized with config file: {self.config_file}")

    def save_config_path(self):
        """
        Save the path to the configuration file in `config_path.txt`.
        """
        if not os.path.isfile(self.config_file):
            raise FileNotFoundError(f"Cannot set config path. File does not exist: '{self.config_file}'")

        test_config_path = os.path.abspath(os.path.join(os.path.dirname(self.config_file), "config_path.txt"))
        config_path_file = test_config_path if self.args.test else CONFIG_PATH_FILE
        with open(config_path_file, "w") as f:
            f.write(self.config_file)
        LOG.info(f"Configuration path saved to '{config_path_file}'.")

    def create_template_config(self):
        """
        Create a template configuration file.
        """
        LOG.info("Creating config ...")

        if self.args.task_server != "celery":
            raise MerlinInvalidTaskServerError("Only celery can be configured currently.")

        template_config = "app_redis.yaml" if self.args.broker == "redis" else "app.yaml"

        with resources.path("merlin.data.celery", template_config) as template_config_file:
            self._create_config(template_config_file)

    def _create_config(self, template_config_file: str):
        """
        Internal method to create the Celery configuration.

        Args:
            template_config_file (str): Path to the template configuration file.
        """
        # Create the configuration file if it doesn't already exist
        if not os.path.isfile(self.config_file):
            os.makedirs(os.path.dirname(self.config_file), exist_ok=True)
            shutil.copy(template_config_file, self.config_file)

            # Check to make sure the copy worked
            if not os.path.isfile(self.config_file):
                LOG.error(f"Cannot create config file '{self.config_file}'.")
            else:
                LOG.info(f"The file '{self.config_file}' is ready to be edited for your system.")
        # Log a message if the configuration file already exists
        else:
            LOG.info(f"The config file already exists, '{self.config_file}'.")

        from merlin.common.security import encrypt  # pylint: disable=import-outside-toplevel

        encrypt.init_key()

    def update_broker(self):
        """
        Update the broker section of the app.yaml file.
        """
        LOG.info(f"Updating broker settings in '{self.config_file}'...")

        with open(self.config_file, "r") as app_yaml_file:
            config = yaml.safe_load(app_yaml_file)

        broker_config = config.get("broker", {})
        if self.args.type == "redis":
            self.update_redis_config(broker_config)
        elif self.args.type == "rabbitmq":
            self.update_rabbitmq_config(broker_config)
        else:
            LOG.error("Invalid broker type. Use 'redis' or 'rabbitmq'.")
            return

        config["broker"] = broker_config

        with open(self.config_file, "w") as app_yaml_file:
            yaml.dump(config, app_yaml_file, default_flow_style=False)

        LOG.info("Broker settings successfully updated. Check your new connection with `merlin info`.")

    def update_backend(self):
        """
        Update the results backend section of the app.yaml file.
        """
        LOG.info(f"Updating results backend settings in '{self.config_file}'...")

        with open(self.config_file, "r") as app_yaml_file:
            config = yaml.safe_load(app_yaml_file)

        backend_config = config.get("results_backend", {})
        if self.args.type == "redis":
            self.update_redis_config(backend_config)
        else:
            LOG.error("Invalid backend type. Use 'redis'.")
            return

        config["results_backend"] = backend_config

        with open(self.config_file, "w") as app_yaml_file:
            yaml.dump(config, app_yaml_file, default_flow_style=False)

        LOG.info("Results backend settings successfully updated. Check your new connection with `merlin info`.")

    def update_redis_config(self, config: Dict[str, Union[str, int]]):
        """
        Update the Redis-specific configuration for either broker or backend.

        Args:
            config: The configuration dictionary to update.
        """
        LOG.warning("Redis does not use the 'username' or 'vhost' arguments. Ignoring these if provided.")
        config["name"] = "rediss"
        config["username"] = ""  # Redis doesn't use a username

        required_fields = ["password_file", "server", "port", "db_num", "cert_reqs"]
        self.update_config(config, required_fields)

    def update_rabbitmq_config(self, config: Dict[str, Union[str, int]]):
        """
        Update the RabbitMQ-specific broker configuration.

        Args:
            config: The configuration dictionary to update.
        """
        LOG.warning("RabbitMQ does not use the 'db_num' argument.")
        config["name"] = "rabbitmq"

        required_fields = ["username", "password_file", "server", "port", "vhost", "cert_reqs"]
        self.update_config(config, required_fields)

    def update_config(self, config: Dict[str, Union[str, int]], required_fields: List[str]):
        """
        Generic function to update configuration fields based on provided arguments.

        Args:
            config: The configuration dictionary to update.
            required_fields: List of required field names for the configuration.
        """
        for field in required_fields:
            value = getattr(self.args, field, None)
            if value is not None:
                if field == "password_file":
                    field = "password"
                    LOG.info("Updating password field.")
                else:
                    try:
                        LOG.info(f"Updating {field} value from '{config[field]}' to '{value}'.")
                    except KeyError:  # This should only be hit in the test suite
                        LOG.info(f"Adding new field '{field}' with value '{value}' to config.")
                config[field] = value if field != "port" else int(value)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Backend factory for selecting and instantiating results backends in Merlin.

This module defines the `MerlinBackendFactory` class, which serves as an abstraction
layer for managing available backend implementations. It supports dynamic selection
and instantiation of backend handlers such as Redis or SQLite, based on user input
or system configuration.

The factory maintains mappings of backend names and aliases, and raises a clear error
if an unsupported backend is requested.
"""

from typing import Dict, List

from merlin.backends.redis.redis_backend import RedisBackend
from merlin.backends.results_backend import ResultsBackend
from merlin.backends.sqlite.sqlite_backend import SQLiteBackend
from merlin.exceptions import BackendNotSupportedError


# TODO add register_backend call to this when we create task server interface?
# TODO could this factory replace the functions in config/results_backend.py?
# - Perhaps it should be a class outside of this?
class MerlinBackendFactory:
    """
    Factory class for managing and instantiating supported Merlin backends.

    This class maintains a registry of available results backends (e.g., Redis, SQLite)
    and provides a unified interface for retrieving a backend implementation by name.

    Attributes:
        _backends (Dict[str, backends.results_backend.ResultsBackend]): Mapping of backend names to their classes.
        _backend_aliases (Dict[str, str]): Optional aliases for resolving canonical backend names.

    Methods:
        get_supported_backends: Returns a list of supported backend names.
        get_backend: Instantiates and returns the backend associated with the given name.
    """

    def __init__(self):
        """Initialize the Merlin backend factory."""
        # Map canonical backend names to their classes
        self._backends: Dict[str, ResultsBackend] = {
            "redis": RedisBackend,
            "sqlite": SQLiteBackend,
        }
        # Map aliases to canonical backend names
        self._backend_aliases: Dict[str, str] = {"rediss": "redis"}

    def get_supported_backends(self) -> List[str]:
        """
        Get a list of the supported backends in Merlin.

        Returns:
            A list of names representing the supported backends in Merlin.
        """
        return list(self._backends.keys())

    def get_backend(self, backend: str) -> ResultsBackend:
        """
        Get backend handler for whichever backend the user is using.

        Args:
            backend: The name of the backend to load up.

        Returns:
            An instantiation of a [`ResultsBackend`][backends.results_backend.ResultsBackend] object.

        Raises:
            (exceptions.BackendNotSupportedError): If the requested backend is not supported.
        """
        # Resolve the alias to the canonical backend name
        backend = self._backend_aliases.get(backend, backend)

        # Get the correct backend class
        backend_object = self._backends.get(backend)

        if backend_object is None:
            raise BackendNotSupportedError(f"Backend unsupported by Merlin: {backend}.")

        return backend_object(backend)


backend_factory = MerlinBackendFactory()



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module houses dataclasses that define the format of the data
that's stored in Merlin's database.
"""

import hashlib
import json
import logging
import os
import uuid
from abc import ABC, abstractmethod
from dataclasses import Field, asdict, dataclass, field
from dataclasses import fields as dataclass_fields
from datetime import datetime
from typing import Dict, List, Set, Tuple, Type, TypeVar

from filelock import FileLock

from merlin.common.enums import WorkerStatus


LOG = logging.getLogger("merlin")
T = TypeVar("T", bound="BaseDataModel")


@dataclass
class BaseDataModel(ABC):
    """
    A base class for dataclasses that provides common serialization, deserialization, and
    update functionality, with support for additional data.

    This class is designed to be extended by other dataclasses and includes methods for
    converting instances to and from dictionaries or JSON, managing fields, and updating
    field values with validation.

    Attributes:
        additional_data: A dictionary to store any extra data not explicitly defined
            as fields in the dataclass.
        fields_allowed_to_be_updated: A list of field names that are allowed to be updated.
            Must be defined in subclasses.

    Methods:
        to_dict:
            Convert the dataclass instance to a dictionary.

        to_json:
            Serialize the dataclass instance to a JSON string.

        from_dict (classmethod):
            Create an instance of the dataclass from a dictionary.

        from_json (classmethod):
            Create an instance of the dataclass from a JSON string.

        dump_to_json_file:
            Dump the data of this dataclass to a JSON file.

        load_from_json_file (classmethod):
            Load the data stored in a JSON file to this dataclass.

        fields:
            Retrieve the fields associated with this dataclass instance or class.

        fields (classmethod):
            Retrieve the fields associated with the dataclass class itself.

        update_fields:
            Update the fields of the dataclass based on a given dictionary of updates.
    """

    additional_data: Dict = field(default_factory=dict)

    def to_dict(self) -> Dict:
        """
        Convert the dataclass to a dictionary.

        Returns:
            The dataclass as a dictionary.
        """
        return asdict(self)

    def to_json(self) -> str:
        """
        Serialize the dataclass to a JSON string.

        Returns:
            The dataclass as a JSON string.
        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls: Type[T], data: Dict) -> T:
        """
        Create an instance of the dataclass from a dictionary.

        Args:
            data: A dictionary to turn into an instance of this dataclass.

        Returns:
            An instance of the dataclass that called this.
        """
        return cls(**data)

    @classmethod
    def from_json(cls: Type[T], json_str: str) -> T:
        """
        Create an instance of the dataclass from a JSON string.

        Args:
            json_str: A JSON string to turn into an instance of this dataclass.

        Returns:
            An instance of the dataclass that called this.
        """
        data = json.loads(json_str)
        return cls.from_dict(data)

    def dump_to_json_file(self, filepath: str):
        """
        Dump the data of this dataclass to a JSON file.

        Args:
            filepath: The path to the JSON file where the data will be written.

        Raises:
            ValueError: If the `filepath` is not provided or is invalid.
        """
        if not filepath:
            raise ValueError("A valid file path must be provided.")

        # Ensure the directory for the file exists
        os.makedirs(os.path.dirname(filepath), exist_ok=True)

        # Create a lock file alongside the target JSON file
        lock_file = f"{filepath}.lock"
        with FileLock(lock_file):  # pylint: disable=abstract-class-instantiated
            # Write the data to the JSON file
            temp_filepath = f"{filepath}.tmp"  # Use a temporary file for atomic writes
            with open(temp_filepath, "w") as json_file:
                json.dump(self.to_dict(), json_file, indent=4)

            # Replace the temporary file with the target file
            os.replace(temp_filepath, filepath)

        LOG.debug(f"Data successfully dumped to {filepath}.")

    @classmethod
    def load_from_json_file(cls: Type[T], filepath: str) -> T:
        """
        Load the data stored in a JSON file to this dataclass.

        Args:
            filepath: The path to the JSON file where the data is located.

        Raises:
            ValueError: If the `filepath` is not provided or is invalid.
        """
        if not filepath or not os.path.exists(filepath):
            raise ValueError("A valid file path must be provided.")

        # Create a lock file alongside the target JSON file
        lock_file = f"{filepath}.lock"
        with FileLock(lock_file):  # pylint: disable=abstract-class-instantiated
            with open(filepath, "r") as json_file:
                # Parse the JSON data into a dictionary
                data = json.load(json_file)

        # Use from_dict to create an instance of the dataclass
        return cls.from_dict(data)

    def get_instance_fields(self) -> Tuple[Field]:
        """
        Get the fields associated with this instance. Added this method so that the dataclass.fields
        doesn't have to be imported each time you want this info.

        Returns:
            A tuple of dataclass.Field objects representing the fields in this data class.
        """
        return dataclass_fields(self)

    @classmethod
    def get_class_fields(cls) -> Tuple[Field]:
        """
        Get the fields associated with this object. Added this method so that the dataclass.fields
        doesn't have to be imported each time you want this info.

        Returns:
            A tuple of dataclass.Field objects representing the fields in this data class.
        """
        return dataclass_fields(cls)

    @property
    @abstractmethod
    def fields_allowed_to_be_updated(self) -> List[str]:
        """
        A property to be overridden in subclasses to define which fields are allowed to be updated.

        Returns:
            A list of fields that are allowed to be updated in this class.
        """

    def update_fields(self, updates: Dict):
        """
        Given a dictionary of updates to be made to this data class, loop through the updates
        applying them when valid.

        Args:
            updates: A dictionary of updates to be made to this data class.
        """
        # Iterate through the updates
        for field_name, new_value in updates.items():
            if field_name == "id":
                continue

            if hasattr(self, field_name):
                if getattr(self, field_name) == new_value:  # Not an update so skip
                    continue

                if field_name in self.fields_allowed_to_be_updated:
                    # Update the allowed field
                    setattr(self, field_name, new_value)
                else:
                    # Log a warning for unauthorized updates
                    LOG.warning(f"Field '{field_name}' is not allowed to be updated. Ignoring the change.")
            else:
                # Log a warning if the field doesn't exist explicitly
                LOG.warning(
                    f"Field '{field_name}' does not explicitly exist in the object. Adding it to the 'additional_data' field."
                )
                self.additional_data[field_name] = new_value


@dataclass
class StudyModel(BaseDataModel):
    """
    A dataclass to store all of the information for a study.

    Attributes:
        additional_data (Dict): For any extra data not explicitly defined.
        fields_allowed_to_be_updated (List[str]): A list of field names that are
            allowed to be updated.
        id (str): The unique ID for the study.
        name (str): The name of the study.
        runs (List[str]): A list of runs associated with this study.
    """

    id: str = field(default_factory=lambda: str(uuid.uuid4()))  # pylint: disable=invalid-name
    name: str = None
    runs: List[str] = field(default_factory=list)

    @property
    def fields_allowed_to_be_updated(self) -> List[str]:
        """
        Define the fields that are allowed to be updated for a `StudyModel` object.

        Returns:
            A list of fields that are allowed to be updated in this class.
        """
        return ["runs"]


@dataclass
class RunModel(BaseDataModel):  # pylint: disable=too-many-instance-attributes
    """
    A dataclass to store all of the information for a run.

    Attributes:
        additional_data (Dict): For any extra data not explicitly defined.
        child (str): The ID of the child run (if any).
        fields_allowed_to_be_updated (List[str]): A list of field names that are allowed
            to be updated.
        id (str): The unique ID for the run.
        parameters (Dict): The parameters used in this run.
        parent (str): The ID of the parent run (if any).
        queues (List[str]): The task queues used for this run.
        run_complete (bool): Wether the run is complete.
        samples (Dict): The samples used in this run.
        steps (List[str]): A list of unique step IDs that are executed in this run.
            Each ID will correspond to a `StepInfo` entry.
        study_id (str): The unique ID of the study this run is associated with.
            Corresponds with a `StudyModel` entry.
        workers (List[str]): A list of worker ids executing tasks for this run. Each ID
            will correspond with a `LogicalWorkerModel` entry.
        workspace (str): The path to the output workspace.
    """

    id: str = field(default_factory=lambda: str(uuid.uuid4()))  # pylint: disable=invalid-name
    study_id: str = None
    workspace: str = None
    steps: List[str] = field(default_factory=list)  # TODO NOT YET IMPLEMENTED
    queues: List[str] = field(default_factory=list)
    workers: List[str] = field(default_factory=list)
    parent: str = None  # TODO NOT YET IMPLEMENTED; do we even have a good way that this and `child` can be set?
    child: str = None  # TODO NOT YET IMPLEMENTED
    run_complete: bool = False
    parameters: Dict = field(default_factory=dict)  # TODO NOT YET IMPLEMENTED
    samples: Dict = field(default_factory=dict)  # TODO NOT YET IMPLEMENTED

    @property
    def fields_allowed_to_be_updated(self) -> List[str]:
        """
        Define the fields that are allowed to be updated for a `RunModel` object.

        Returns:
            A list of fields that are allowed to be updated in this class.
        """
        return ["parent", "child", "run_complete", "additional_data", "workers"]


@dataclass
class LogicalWorkerModel(BaseDataModel):
    """
    Represents a high-level definition of a Celery worker, as defined by the user.

    Logical workers are abstract representations of workers that define their behavior
    and configuration, such as the queues they listen to and their name. They are unique
    based on their name and queues, and do not correspond directly to any running process.
    Instead, they serve as templates or logical definitions from which physical workers
    are created.

    Note:
        Logical workers are abstract and do not represent actual running processes. They are
        used to define worker behavior and configuration at a high level, while physical workers
        represent the actual running instances of these logical definitions.

    Attributes:
        additional_data (Dict): For any extra data not explicitly defined.
        fields_allowed_to_be_updated (List[str]): A list of field names that are
            allowed to be updated.
        id (str): A unique identifier for the logical worker. Defaults to a UUID string.
        name (str): The name of the logical worker.
        physical_workers (List[str]): A list of unique IDs of the physical worker instances
            created from this logical instance. Corresponds with
            [`PhyiscalWorkerModel`][db_scripts.data_models.PhysicalWorkerModel] entries.
        queues (List[str]): A list of task queues the worker is listening to.
        runs (List[str]): A list of unique IDs of the runs using this worker.
            Corresponds with [`RunModel`][db_scripts.data_models.RunModel] entries.
    """

    name: str = None
    queues: Set[str] = field(default_factory=set)
    id: str = None  # pylint: disable=invalid-name
    runs: List[str] = field(default_factory=list)
    physical_workers: List[str] = field(default_factory=list)

    def __post_init__(self):
        """
        Generate and save a UUID based on the values of `name` and `queues`, to help ensure that
        each logical worker is unique to these values.

        Raises:
            TypeError: When name or queues are not provided to the constructor.
        """
        if self.name is None or not self.queues:
            raise TypeError("The `name` and `queues` arguments of LogicalWorkerModel are required.")

        generated_id = self.generate_id(self.name, self.queues)
        if self.id != generated_id:
            if self.id is not None:
                LOG.warning(f"ID '{self.id}' for LogicalWorkerModel was provided but it will be overwritten.")
            self.id = generated_id

    @classmethod
    def generate_id(cls, name: str, queues: List[str]) -> uuid.UUID:
        """
        Generate a UUID based on the values of `name` and `queues`.

        Args:
            name: The name of the logical worker.
            queues: The queues that the logical worker is assigned.

        Returns:
            A UUID based on the values of `name` and `queues`.
        """
        unique_string = f"{name}:{','.join(sorted(queues))}"
        hex_string = hashlib.md5(unique_string.encode("UTF-8")).hexdigest()
        return str(uuid.UUID(hex=hex_string))

    @property
    def fields_allowed_to_be_updated(self) -> List[str]:
        """
        Define the fields that are allowed to be updated for a `LogicalWorkerModel` object.

        Returns:
            A list of fields that are allowed to be updated in this class.
        """
        return ["runs", "physical_workers"]


@dataclass
class PhysicalWorkerModel(BaseDataModel):  # pylint: disable=too-many-instance-attributes
    """
    Represents a running instance of a Celery worker, created from a logical worker definition.

    Physical workers are the actual implementations of logical workers, running as processes on a host machine.
    They are responsible for executing tasks defined in the queues specified by their corresponding logical worker.
    Each physical worker is uniquely identified and includes runtime-specific details such as its PID, status, and
    heartbeat timestamp.

    Attributes:
        additional_data (Dict): For any extra data not explicitly defined.
        args (Dict): A dictionary of arguments used to configure the worker.
        fields_allowed_to_be_updated (List[str]): A list of field names that are
            allowed to be updated.
        heartbeat_timestamp (datetime): The last time the worker sent a heartbeat signal.
        host (str): The hostname or IP address of the machine running the worker.
        id (str): A unique identifier for the physical worker. Defaults to a UUID string.
        latest_start_time (datetime): The timestamp when the worker process was last started.
        launch_cmd (str): The command used to launch the worker process.
        logical_worker_id (str): The ID of the logical worker that this was created from.
        name (str): The name of the physical worker.
        pid (str): The process ID (PID) of the worker process.
        restart_count (int): The number of times this worker has been restarted.
        status (WorkerStatus): The current status of the worker (e.g., running, stopped).
    """

    id: str = field(default_factory=lambda: str(uuid.uuid4()))  # pylint: disable=invalid-name
    logical_worker_id: str = None
    name: str = None  # Will be of the form celery@worker_name.hostname
    launch_cmd: str = None
    args: Dict = field(default_factory=dict)
    pid: str = None
    status: WorkerStatus = WorkerStatus.STOPPED
    heartbeat_timestamp: datetime = field(default_factory=datetime.now)
    latest_start_time: datetime = field(default_factory=datetime.now)
    host: str = None
    restart_count: int = 0

    @property
    def fields_allowed_to_be_updated(self) -> List[str]:
        """
        Define the fields that are allowed to be updated for a `PhysicalWorkerModel` object.

        Returns:
            A list of fields that are allowed to be updated in this class.
        """
        return [
            "launch_cmd",
            "args",
            "pid",
            "status",
            "heartbeat_timestamp",
            "latest_start_time",
            "restart_count",
        ]


# TODO create a StepInfo class to store information about a step
# - Can probably link this to status
# - Each step should have entries for parameters/samples but only those that are actually used in the step
